/*
 * $Id: libgnac-converter.c 379 2009-07-22 17:35:09Z bdupasqu $
 *
 * This file is part of GNAC - Gnome Audio Converter
 *
 * Copyright (C) 2007 - 2009 Gnac
 *    
 *    - DUPASQUIER  Benoit    <bdupasqu@users.sourceforge.net>
 *    - JOAQUIM     David     <daiji@users.sourceforge.net>
 *    - ROUX        Alexandre <alexandreroux@users.sourceforge.net>
 *
 * GNAC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GNAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNAC; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <glib/gi18n.h>

#include "libgnac-converter.h"
#include "libgnac-debug.h"
#include "libgnac-error.h"
#include "libgnac-gst.h"
#include "libgnac-marshallers.h"
#include "libgnac-output.h"

enum 
{
  OVERWRITE,
  PROGRESS,
  FILE_ADDED,
  FILE_REMOVED,
  FILE_STARTED,
  FILE_COMPLETED,
  FILES_CLEARED,
  STARTED,
  STOPPED,
  COMPLETION,
  ERROR,
  LAST_SIGNAL
};

enum
{
  PROP_0, 
  PROP_PIPELINE_DESC,
  PROP_STRIP_SPECIAL,
  PROP_RENAME_PATTERN,
  PROP_FOLDER_HIERARCHY,
  PROP_FOLDER_TYPE,
  PROP_FOLDER_PATH,
  PROP_EXTENSION,
  PROP_NB_FILES
};

static guint signals[LAST_SIGNAL] = { 0 };

struct LibgnacConverterPrivate
{
  GHashTable           *file_table;
  GQueue               *queue;
  GQueue               *queue_copy;
  //GHashTableIter        iter;
  LibgnacConverterItem *current;
  gint                  n_converted;
  guint64               elapsed_duration;
  guint64               total_duration;

  gboolean              strip_special;
  gchar                *pipeline_description;
  gchar                *rename_pattern;
  gint                  folder_type;
  gchar                *folder_hierarchy;
  gchar                *folder_path;
  gchar                *extension;
};

G_DEFINE_TYPE(LibgnacConverter, libgnac_converter, G_TYPE_OBJECT);

#define LIBGNAC_CONVERTER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), \
    LIBGNAC_TYPE_CONVERTER, LibgnacConverterPrivate))


LibgnacMetadata *metadata;


static void
libgnac_converter_init(LibgnacConverter *self)
{
  LibgnacConverterPrivate *priv;

  self->priv = priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  /* Public */

  /* Private */

  priv->queue = g_queue_new();
  priv->queue_copy = NULL;

  priv->file_table = g_hash_table_new_full(g_str_hash, 
        g_str_equal, 
        g_free, 
        (GDestroyNotify)libgnac_converter_item_free);

  priv->current = NULL;
  priv->n_converted = -1;
  priv->total_duration = 0;
  priv->elapsed_duration = 0;

  priv->pipeline_description = NULL;
  priv->folder_path = NULL;
  priv->extension = NULL;
  priv->rename_pattern = NULL;
  priv->folder_hierarchy = NULL;
  priv->folder_type = 0;
}


static void
libgnac_converter_dispose(GObject *gobject)
{
  LibgnacConverter *self = LIBGNAC_CONVERTER(gobject);

  if (self->priv->file_table)
  {
    g_hash_table_unref(self->priv->file_table);
    self->priv->file_table = NULL;
  }

  if (self->priv->queue)
  {
    g_queue_free(self->priv->queue);
    self->priv->queue = NULL;
  }

  /* Chain up to the parent class */
  G_OBJECT_CLASS(libgnac_converter_parent_class)->dispose(gobject);
}


static void
libgnac_converter_finalize(GObject *gobject)
{
  LibgnacConverter *self = LIBGNAC_CONVERTER(gobject);

  if (self->priv->pipeline_description) {
    g_free(self->priv->pipeline_description);
  }

  if (self->priv->extension) {
    g_free(self->priv->extension);
  }

  if (self->priv->folder_path) {
    g_free(self->priv->folder_path);
  }

  libgnac_output_finalize();

  /* Chain up to the parent class */
  G_OBJECT_CLASS(libgnac_converter_parent_class)->finalize(gobject);
}


static void
libgnac_converter_get_property(GObject    *object,
                               guint       property_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
  LibgnacConverter *self = LIBGNAC_CONVERTER(object);

  switch (property_id) 
  {
    case PROP_PIPELINE_DESC:
      g_value_set_string(value, self->priv->pipeline_description); 
    break;

    case PROP_STRIP_SPECIAL:
      g_value_set_boolean(value, self->priv->strip_special);
    break;

    case PROP_RENAME_PATTERN:
      g_value_set_string(value, self->priv->rename_pattern); 
    break;

    case PROP_FOLDER_HIERARCHY:
      g_value_set_string(value, self->priv->folder_hierarchy); 
    break;

    case PROP_FOLDER_TYPE:
      g_value_set_int(value, self->priv->folder_type); 
    break;

    case PROP_FOLDER_PATH:
      g_value_set_string(value, self->priv->folder_path); 
    break;

    case PROP_EXTENSION:
      g_value_set_string(value, self->priv->extension); 
    break;

    case PROP_NB_FILES:
      g_value_set_int(value, g_hash_table_size(self->priv->file_table));
    break;

    default:
      /* We don't have any other property... */
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }
}


static void
libgnac_converter_set_property(GObject      *object,
                               guint         property_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
  LibgnacConverter *self = LIBGNAC_CONVERTER(object);

  switch (property_id) 
  {
    case PROP_PIPELINE_DESC:
      g_free(self->priv->pipeline_description);
      self->priv->pipeline_description = NULL;
      self->priv->pipeline_description = g_value_dup_string(value);
    break;

    case PROP_STRIP_SPECIAL:
      self->priv->strip_special = g_value_get_boolean(value);
    break;

    case PROP_RENAME_PATTERN:
      self->priv->rename_pattern = g_value_dup_string(value);
    break;

    case PROP_FOLDER_HIERARCHY:
      g_free(self->priv->folder_hierarchy);
      self->priv->folder_hierarchy = g_value_dup_string(value);
    break;

    case PROP_FOLDER_TYPE:
      self->priv->folder_type = g_value_get_int(value);
    break;

    case PROP_FOLDER_PATH:
      g_free(self->priv->folder_path);
      self->priv->folder_path = g_value_dup_string(value);
    break;

    case PROP_EXTENSION:
      g_free(self->priv->extension);
      self->priv->extension = NULL;
      self->priv->extension = g_value_dup_string(value);
    break;

    default:
      /* We don't have any other property... */
      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }
}


static gboolean
libgnac_converter_allow_overwrite_accumulator(GSignalInvocationHint *ihint,
                                             GValue                *return_accu,
                                             const GValue          *handler_return,
                                             gpointer data)
{
  gboolean allow_overwrite;

  allow_overwrite = g_value_get_boolean(handler_return);
  g_value_set_boolean(return_accu, allow_overwrite);

  return FALSE;
}


static void
libgnac_converter_class_init(LibgnacConverterClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
  GParamSpec *pspec;

  g_type_class_add_private(klass, sizeof(LibgnacConverterPrivate));

  gobject_class->dispose = libgnac_converter_dispose;
  gobject_class->finalize = libgnac_converter_finalize;

  gobject_class->set_property = libgnac_converter_set_property;
  gobject_class->get_property = libgnac_converter_get_property;

  /* Pipeline desc */
  pspec = g_param_spec_string("pipeline-description",
                              "Pipeline Description",
                              "Pipeline output description",
                              NULL,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_PIPELINE_DESC,
                                  pspec);

  /* Folder hierarchy */
  pspec = g_param_spec_string("folder-hierarchy",
                              "Folder Hierarchy",
                              "Folder hierarchy to create",
                              NULL,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_FOLDER_HIERARCHY,
                                  pspec);

  /* Folder path */
  pspec = g_param_spec_string("folder-path",
                              "Folder Path",
                              "Path to the output folder",
                              NULL,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_FOLDER_PATH,
                                  pspec);

  /* Extension */
  pspec = g_param_spec_string("extension",
                              "File Extension",
                              "Output's file extension",
                              NULL,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_EXTENSION,
                                  pspec);

  /* Strip special characters */
  pspec = g_param_spec_boolean("strip-special",
                              "Strip special",
                              "Strip special characters",
                              FALSE,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_STRIP_SPECIAL,
                                  pspec);

  /* Rename pattern */
  pspec = g_param_spec_string("rename-pattern",
                              "Rename Pattern",
                              "Output's rename pattern",
                              NULL,
                              G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_RENAME_PATTERN,
                                  pspec);

  /* Folder type */
  pspec = g_param_spec_int("folder-type",
                           "Folder type",
                           "Output's folder type",
                           0, FOLDER_TYPE_N, 0,
                           G_PARAM_READWRITE);

  g_object_class_install_property(gobject_class, 
                                  PROP_FOLDER_TYPE,
                                  pspec);

  /* Nb files */
  pspec = g_param_spec_int("nb-files",
                           "File number",
                           "Output's nb files",
                           0, G_MAXINT, 0,
                           G_PARAM_READABLE);

  g_object_class_install_property(gobject_class, 
                                  PROP_NB_FILES,
                                  pspec);


  /* Signals */

  signals[OVERWRITE] = g_signal_new("allow-overwrite",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, overwrite),
      libgnac_converter_allow_overwrite_accumulator, NULL,
      g_cclosure_user_marshal_BOOLEAN__POINTER,
      G_TYPE_BOOLEAN, 1, G_TYPE_POINTER);

  signals[PROGRESS] = g_signal_new("progress",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, progress),
      NULL, NULL,
      g_cclosure_user_marshal_VOID__UINT_FLOAT_UINT64,
      G_TYPE_NONE, 3, G_TYPE_UINT, G_TYPE_FLOAT, G_TYPE_UINT64);

  signals[FILE_ADDED] = g_signal_new("file-added",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, file_added),
      NULL, NULL,
      g_cclosure_marshal_VOID__STRING,
      G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[FILE_REMOVED] = g_signal_new("file-removed",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, file_removed),
      NULL, NULL,
      g_cclosure_marshal_VOID__STRING,
      G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[FILE_STARTED] = g_signal_new("file-started",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, file_started),
      NULL, NULL,
      g_cclosure_marshal_VOID__STRING,
      G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[FILE_COMPLETED] = g_signal_new("file-completed",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, file_completed),
      NULL, NULL,
      g_cclosure_marshal_VOID__STRING,
      G_TYPE_NONE, 1, G_TYPE_STRING);

  signals[FILES_CLEARED] = g_signal_new("files-cleared",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, files_cleared),
      NULL, NULL,
      g_cclosure_marshal_VOID__VOID,
      G_TYPE_NONE, 0);

  signals[STARTED] = g_signal_new("started",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, started),
      NULL, NULL,
      g_cclosure_marshal_VOID__VOID,
      G_TYPE_NONE, 0);

  signals[STOPPED] = g_signal_new("stopped",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, stopped),
      NULL, NULL,
      g_cclosure_marshal_VOID__VOID,
      G_TYPE_NONE, 0);

  signals[COMPLETION] = g_signal_new("completion",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, completion),
      NULL, NULL,
      g_cclosure_marshal_VOID__VOID,
      G_TYPE_NONE, 0);

  signals[ERROR] = g_signal_new("error",
      G_TYPE_FROM_CLASS(gobject_class),
      G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET(LibgnacConverterClass, error),
      NULL, NULL,
      g_cclosure_user_marshal_VOID__STRING_STRING_POINTER,
      G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
}


/* private methods */

static gboolean
libgnac_converter_start_next(LibgnacConverter *self)
{
  LibgnacConverterPrivate *priv;
  GError *err = NULL;
  LibgnacConverterItem *item = NULL;
  gchar *key = NULL;
  gchar *uri = NULL;

  g_return_val_if_fail(LIBGNAC_IS_CONVERTER(self), FALSE);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  do {
    LibgnacTags *tags;

    key = g_queue_pop_head(priv->queue_copy);
    if (key) item = g_hash_table_lookup(priv->file_table, key);
    else item = NULL;
    
    if (!item) 
    {
      return FALSE;
    }

    if (err) {
      g_clear_error(&err);
    } 

    priv->n_converted++;

    priv->current = item;

    uri = g_file_get_uri(item->source);

    g_signal_emit(self, signals[FILE_STARTED], 0, uri);
    libgnac_converter_item_run(item, &err);
    if (err)  {
      g_signal_emit(item->converter, signals[ERROR], 0, uri, err->message, err);
    }

    tags = libgnac_metadata_extract(metadata, item->source, NULL);
    if (libgnac_metadata_tags_exist(tags, GST_TAG_DURATION, NULL))
    {
      priv->elapsed_duration += g_value_get_uint64(
                      LIBGNAC_METADATA_TAG_DURATION(tags));
    }

    g_free(uri);

  } while (err);

  return TRUE;
}


static gboolean
libgnac_converter_restart_current(LibgnacConverter *self)
{
  LibgnacConverterPrivate *priv;
  GError *err = NULL;

  g_return_val_if_fail(LIBGNAC_IS_CONVERTER(self), FALSE);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  libgnac_debug("restarting pipeline");

  if (priv->current) 
  {
      libgnac_converter_item_run(priv->current, &err);
      if (err) {
        g_clear_error(&err);
        return FALSE;
      }
      return TRUE;
  }
  return FALSE;
}


/* public methods */

GObject *
libgnac_converter_new(LibgnacMetadata *mdata)
{
  metadata = mdata;
  return g_object_new(LIBGNAC_TYPE_CONVERTER, NULL);
}


void
libgnac_converter_add(LibgnacConverter  *self, 
                      GFile             *file, 
                      GError           **error)
{
  LibgnacConverterPrivate *priv;
  LibgnacConverterItem *item;
  gchar *uri;
  gchar *dup_uri;


  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  uri = g_file_get_uri(file);
  dup_uri = g_strdup(uri);

  if (!g_hash_table_lookup(priv->file_table, uri)) 
  {
    LibgnacTags *tags;

    item = libgnac_converter_item_new(self, file);
    // dup uri is freed within the hashtable
    g_queue_push_head(priv->queue, dup_uri);
    g_hash_table_insert(priv->file_table, dup_uri, item);
    g_signal_emit(self, signals[FILE_ADDED], 0, uri);

    tags = libgnac_metadata_extract(metadata, item->source, NULL);
    if (libgnac_metadata_tags_exist(tags, GST_TAG_DURATION, NULL))
    {
      priv->total_duration += g_value_get_uint64(
                      LIBGNAC_METADATA_TAG_DURATION(tags));
    }
  } 
  else 
  {
   g_set_error(error,
     LIBGNAC_ERROR,
     LIBGNAC_ERROR_SAME_FILE,
     _("File %s is already in list"),
     uri);
  }

  g_free(uri);
}


void
libgnac_converter_remove(LibgnacConverter  *self, 
                         GFile             *file, 
                         GError           **error)
{
  LibgnacConverterPrivate *priv;
  gchar                   *uri;
  gpointer                 key_uri = NULL;
  gpointer                 item = NULL;
  
  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  uri = g_file_get_uri(file);

  if (g_hash_table_lookup_extended(priv->file_table, uri, &key_uri, &item))
  {
    GList *elem;
    LibgnacTags *tags;
    
    elem = g_queue_find(priv->queue, key_uri);

    g_return_if_fail(elem);
   
    tags = libgnac_metadata_extract(metadata, 
                    ((LibgnacConverterItem*)item)->source, NULL);
    if (libgnac_metadata_tags_exist(tags, GST_TAG_DURATION, NULL))
    {
      priv->total_duration -= g_value_get_uint64(
                    LIBGNAC_METADATA_TAG_DURATION(tags));
    }

    g_queue_remove(priv->queue, elem->data);
    g_hash_table_remove(priv->file_table, uri);

    g_signal_emit(self, signals[FILE_REMOVED], 0, uri);
  } 
  else 
  {
    libgnac_debug("item not found for file %s", uri);
    g_set_error(error,  
      LIBGNAC_ERROR, 
      LIBGNAC_ERROR_MISSING_FILE, 
      _("File %s isn't in list"),
      uri);
  }

  g_free(uri);
}


void
libgnac_converter_clear(LibgnacConverter *self)
{
  LibgnacConverterPrivate *priv;
  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  g_queue_clear(priv->queue);
  g_hash_table_remove_all(priv->file_table);

  priv->total_duration = 0;

  g_signal_emit(self, signals[FILES_CLEARED], 0);
}


void
libgnac_converter_start(LibgnacConverter  *self, 
                        GError           **error)
{
  LibgnacConverterPrivate *priv;
  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  if (priv->queue_copy) {
    g_queue_free(priv->queue_copy);
    priv->queue_copy = NULL;
  }
  priv->queue_copy = g_queue_copy(priv->queue);
  g_queue_sort(priv->queue_copy, (GCompareDataFunc)g_strcmp0, NULL);
  //g_hash_table_iter_init(&(priv->iter), priv->file_table);
  priv->n_converted = -1;
  priv->elapsed_duration = 0;

  g_signal_emit(self, signals[STARTED], 0);

  if (!libgnac_converter_start_next(self))
  {
    g_signal_emit(self, signals[COMPLETION], 0);
  }
}


void
libgnac_converter_stop(LibgnacConverter  *self, 
                       GError           **error)
{
  GError *err = NULL;
  LibgnacConverterPrivate *priv;

  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);

  libgnac_converter_item_stop(priv->current, &err);
  if (err) {
    g_propagate_error(error, err);
    return;
  }
  priv->elapsed_duration = 0;
  priv->n_converted = -1;
  priv->current = NULL;
  g_signal_emit(self, signals[STOPPED], 0);
}


void
libgnac_converter_pause(LibgnacConverter  *self,
                        GError           **error)
{
  LibgnacConverterPrivate *priv;
  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);
  libgnac_gst_pause(priv->current, error);
}


void
libgnac_converter_resume(LibgnacConverter  *self, 
                         GError           **error)
{
  LibgnacConverterPrivate *priv;
  g_return_if_fail(LIBGNAC_IS_CONVERTER(self));
  g_return_if_fail(error == NULL || *error == NULL);

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(self);
  libgnac_gst_run(priv->current, error);
}


/* LibgnacConverterItem functions */

LibgnacConverterItem *
libgnac_converter_item_new(LibgnacConverter *self, 
                           GFile            *source)
{
  LibgnacConverterItem *item;

  g_return_val_if_fail(LIBGNAC_IS_CONVERTER(self), NULL);

  item = (LibgnacConverterItem*)g_malloc(sizeof(LibgnacConverterItem));
  
  g_object_ref(self);
  item->converter = self;
  item->source = g_file_dup(source);
  item->destination = NULL;
  item->pipeline = NULL;
  item->audio_encoder = NULL;
  item->bus = NULL;
  item->timeout_id = 0;
  item->pos = 0;

  return item;
}


void
libgnac_converter_item_build(LibgnacConverterItem  *item, 
                             GError               **error)
{
  GError *err = NULL;
  g_return_if_fail(error == NULL || *error == NULL);


  libgnac_gst_build_pipeline(item, &err);
  if (err) {
    libgnac_debug("Unable to build pipeline");
    libgnac_gst_clean_pipeline(item);
    g_propagate_error(error, err);
  }
}


void
libgnac_converter_item_run(LibgnacConverterItem  *item, 
                           GError               **error)
{
  GError *err = NULL;
  g_return_if_fail(error == NULL || *error == NULL);

  /* Build pipeline if doesn't exist */
  if (!item->pipeline) {
    libgnac_converter_item_build(item, &err);
    if(err)
    {
      g_propagate_error(error, err);
      return;
    }
  }

  libgnac_converter_item_initialize_event_source(item);
  
  libgnac_gst_run(item, &err);
  if (err) {
    libgnac_debug("Unable to run pipeline");
    libgnac_gst_clean_pipeline(item);
    g_propagate_error(error, err);
    return;
  }
}


void
libgnac_converter_item_stop(LibgnacConverterItem  *item, 
                            GError               **error)
{
  GError *err = NULL;
  g_return_if_fail(error == NULL || *error == NULL);

  libgnac_converter_item_clear_event_source(item);
  libgnac_gst_stop(item, &err);
  // TODO propagate error here
  //g_propagate_error(error, err);

  /* Remove not completed file */
  if (G_IS_FILE(item->destination)) 
  {
    g_file_delete(item->destination, NULL, &err);
    if (err) {
      // Warning, but not critical
      libgnac_debug("Unable to remove partial file");
      g_clear_error(&err);
    }
  }
}


/*void
libgnac_converter_item_pause(LibgnacConverterItem  *item, 
                             GError               **error)
{
  g_return_if_fail(error == NULL || *error == NULL);
}


void
libgnac_converter_item_resume(LibgnacConverterItem  *item, 
                              GError               **error)
{
  g_return_if_fail(error == NULL || *error == NULL);
}*/


void
libgnac_converter_item_free(LibgnacConverterItem *item)
{
  libgnac_converter_item_clear_event_source(item);
  g_object_unref(item->converter);
  item->converter = NULL;

  g_object_unref(item->source);
  item->source = NULL;
  if (item->destination) {
    g_object_unref(item->destination);
    item->destination = NULL;
  }

  libgnac_gst_clean_pipeline(item);

  g_free(item);
}


void
libgnac_converter_item_initialize_event_source(LibgnacConverterItem *item)
{
  item->timeout_id = g_timeout_add(PROGRESS_TIMEOUT, 
      (GSourceFunc)libgnac_converter_percentage_cb, item);
}


void
libgnac_converter_item_clear_event_source(LibgnacConverterItem *item)
{
  if (item->timeout_id) {
    g_source_remove(item->timeout_id);
    item->timeout_id = 0;
  }
}


/* Callbacks */

void
libgnac_converter_eos_cb(GstBus     *bus, 
                         GstMessage *message, 
                         gpointer    data)
{
  LibgnacConverterItem *item;
  gchar *uri;

  item = (LibgnacConverterItem*)data;
  libgnac_converter_item_clear_event_source(item);

  uri = g_file_get_uri(item->source);

  g_signal_emit(item->converter, signals[FILE_COMPLETED], 0, uri);

  /* Start next file conversion */
  if (!libgnac_converter_start_next(item->converter))
  {
    item->converter->priv->n_converted = -1;
    g_signal_emit(item->converter, signals[COMPLETION], 0);
  }
}


void
libgnac_converter_error_cb(GstBus      *bus, 
                           GstMessage  *message, 
                           gpointer     data)
{
  GError *error = NULL;
  GstObject *src = NULL;
  LibgnacConverterItem *item = NULL;
  gchar *uri;
  gchar *name;

  item = (LibgnacConverterItem*)data;
  libgnac_converter_item_clear_event_source(item);

  src = GST_MESSAGE_SRC(message);
  name = gst_object_get_name(src);

  uri = g_file_get_uri(item->source);

  gst_message_parse_error(message, &error, NULL);

  if (g_str_equal("sink", name)) 
  {
    if (g_error_matches(error, 
            GST_RESOURCE_ERROR, 
            GST_RESOURCE_ERROR_NOT_FOUND)) 
    {
      /* Create missing directory */
      GError *err = NULL;
      GFile *folder;

      folder = g_file_get_parent(item->destination);
      g_file_make_directory_with_parents(folder, NULL, &err);
      g_object_unref(folder);
      if (err) {
        g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                        _("Unable to create destination directory"), NULL);
        g_clear_error(&err);
      } else if (libgnac_converter_restart_current(item->converter)) return;

    } else if (g_error_matches(error,
                  GST_RESOURCE_ERROR,
                  GST_RESOURCE_ERROR_OPEN_READ))
    {
      if (g_file_test(g_file_get_path(item->destination), G_FILE_TEST_EXISTS))
      {
        gboolean overwrite = FALSE;
        g_signal_emit(item->converter, signals[OVERWRITE], 0, 
                        item->destination, &overwrite);
        if (overwrite)
        {
          GError *err = NULL;
          g_file_delete(item->destination, NULL, &err);
          if (err) {
            libgnac_debug("Unable to overwrite file %s: %s", uri, err->message);
            g_clear_error(&err);
          }
          if (!err && libgnac_converter_restart_current(item->converter)) 
          {
            g_clear_error(&error);
            g_free(name);
            g_free(uri);
            return;
          }
        }
        else 
        {
          GError *err = NULL;
          g_set_error(&err, LIBGNAC_ERROR, LIBGNAC_ERROR_FILE_EXISTS,
                        _("Destination file %s already exists"), uri);
          g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                        _("Destination file already exists"), err);
          g_clear_error(&err);
        }
      } 
      else
      {
        g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                          _("Unable to access destination file"), error);
      }
    }
  } else if (g_str_equal("source", name)) {
    g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                    _("Unable to read source file"), error);
  } else {
    GError *err = NULL;
    g_file_delete(item->destination, NULL, &err);
    if (err) {
      libgnac_debug("Unable to remove partial file %s: %s", uri, err->message);
      g_clear_error(&err);
    }
    g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                    _("An error occured during conversion"), error);
  }

  libgnac_debug("Conversion failed:\n Object: %s\n Message: %s\n File: %s", 
                  name, error->message, uri);

  g_clear_error(&error);
  g_free(name);
  g_free(uri);

  if (!libgnac_converter_start_next(item->converter))
  {
    item->converter->priv->n_converted = -1;
    g_signal_emit(item->converter, signals[COMPLETION], 0);
  }     

  //libgnac_converter_restart_current(item->converter);
}


void
libgnac_converter_missing_plugin_result_cb(GstInstallPluginsReturn result,
                                           gpointer                data)
{
  LibgnacConverterItem *item;
  gchar *uri;

  item = (LibgnacConverterItem*)data;

  if (result == GST_INSTALL_PLUGINS_SUCCESS)
  {
    if (libgnac_converter_restart_current(item->converter))
    {
      return;
    }
  } else {
    uri = g_file_get_uri(item->source);
    g_signal_emit(item->converter, signals[ERROR], 0, uri, 
                    _("unable to handle this format"), NULL);
    g_free(uri);
  }

  if (!libgnac_converter_start_next(item->converter))
  {
    item->converter->priv->n_converted = -1;
    g_signal_emit(item->converter, signals[COMPLETION], 0);
  }     
}


gboolean
libgnac_converter_percentage_cb(LibgnacConverterItem *item)
{
  gint64      pos = 0;
  gint64      len = 0;
  guint64     time_left = 0;
  GstState    state;
  GstState    pending_state;
  gfloat      result = 0.0;
  LibgnacConverterPrivate *priv;

  static GstFormat format = GST_FORMAT_TIME;

  priv = LIBGNAC_CONVERTER_GET_PRIVATE(item->converter);

  gst_element_get_state(item->pipeline, &state, &pending_state, 0);

  if (state == GST_STATE_PAUSED) 
  {
    return TRUE;
  }
  else if (state != GST_STATE_PLAYING && pending_state != GST_STATE_PLAYING) 
  {
    item->timeout_id = 0;
    return FALSE;
  }

  if (gst_element_query_position(item->pipeline, &format, &pos)) 
  {
    if (pos != item->pos) 
    {
      if (gst_element_query_duration(item->pipeline, &format, &len))
      {
        gfloat processed = priv->elapsed_duration + ((pos - len) / GST_SECOND);
        result = CLAMP((gfloat) processed / priv->total_duration, 0.0f, 1.0f);
        /* XXX this is a workaround to prevent the remaining time
         * to get crazy when we have to deal with mp3wraped files */
        if (priv->total_duration > processed) 
                time_left = priv->total_duration - processed;
      }
      g_signal_emit(item->converter, signals[PROGRESS], 0, 
                      priv->n_converted+1, result, time_left);
    }
  }

  return TRUE;
}
