/*
 * $Id: libgnac-metadata.c 376 2009-07-20 07:14:12Z bdupasqu $
 *
 * This file is part of GNAC - Gnome Audio Converter
 *
 * Copyright (C) 2007 - 2009 Gnac
 *    
 *    - DUPASQUIER  Benoit    <bdupasqu@users.sourceforge.net>
 *    - JOAQUIM     David     <daiji@users.sourceforge.net>
 *    - ROUX        Alexandre <alexandreroux@users.sourceforge.net>
 *
 * GNAC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GNAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNAC; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glib/gi18n.h>
#include <gst/gst.h>
#include <gst/tag/tag.h>
#include <stdarg.h>

#include "libgnac-debug.h"
#include "libgnac-error.h"
#include "libgnac-metadata.h"

static LibgnacTags *dummy_tags;

struct LibgnacMetadataPrivate
{
  GFile *uri;

  GstElement *pipeline;
  GstElement *sink;
  
  GHashTable *tags_table;

  LibgnacTags *metadata;

  gboolean eos;
  GError *error;
};

G_DEFINE_TYPE(LibgnacMetadata, libgnac_metadata, G_TYPE_OBJECT);

#define LIBGNAC_METADATA_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), \
    LIBGNAC_TYPE_METADATA, LibgnacMetadataPrivate))


static void
libgnac_metadata_destroy_pipeline(LibgnacMetadata *md)
{
  if (md->priv->pipeline) 
  {
    GstBus *bus;
    bus = gst_element_get_bus(md->priv->pipeline);
    gst_element_set_state(md->priv->pipeline, GST_STATE_NULL);
    gst_bus_set_sync_handler(bus, NULL, NULL);
    gst_object_unref(GST_OBJECT(bus));
    gst_object_unref(GST_OBJECT(md->priv->pipeline));
    md->priv->pipeline = NULL;
  } 
}


static void
libgnac_metadata_reset(LibgnacMetadata *md)
{
  if (md->priv->uri) {
    g_object_unref(md->priv->uri);
  }
  md->priv->uri = NULL;
  md->priv->eos = FALSE;
  md->priv->error = NULL;
  libgnac_metadata_destroy_pipeline(md);
  /* must be freed by the user */
  if (md->priv->metadata) {
    md->priv->metadata = NULL;
  }
}


static void
libgnac_metadata_dispose(GObject *gobject) 
{
  LibgnacMetadata *self = LIBGNAC_METADATA(gobject);

  if (self->priv->tags_table)
  {
    g_hash_table_destroy(self->priv->tags_table);
    self->priv->tags_table = NULL;
  }

  G_OBJECT_CLASS(libgnac_metadata_parent_class)->dispose(gobject);
}


static void
libgnac_metadata_finalize(GObject *object)
{
  LibgnacMetadata *md = LIBGNAC_METADATA(object);

  libgnac_metadata_reset(md);
  
  G_OBJECT_CLASS(libgnac_metadata_parent_class)->finalize(object);
}


static void
libgnac_metadata_class_init(LibgnacMetadataClass *klass)
{
  GObjectClass *object_class;
  object_class = G_OBJECT_CLASS(klass);

  g_type_class_add_private(klass, sizeof(LibgnacMetadataPrivate));

  object_class->finalize = libgnac_metadata_finalize;
  object_class->dispose = libgnac_metadata_dispose;
}


static void
libgnac_metadata_init(LibgnacMetadata *md)
{
  LibgnacMetadataPrivate *priv = md->priv = LIBGNAC_METADATA_GET_PRIVATE(md);

  priv->tags_table = g_hash_table_new_full(g_str_hash, 
        g_str_equal, 
        g_free, 
        (GDestroyNotify)libgnac_metadata_tags_free);
}


LibgnacMetadata *
libgnac_metadata_new(void)
{
  return LIBGNAC_METADATA(g_object_new(LIBGNAC_TYPE_METADATA, NULL, NULL));
}


static void
libgnac_metadata_free_gvalue(GValue *value) 
{
  if (value) {
    g_value_unset(value);
    g_free(value);
  }   
}


static void
libgnac_metadata_size_prepared_cb(GdkPixbufLoader *loader,
                                  gint             width,
                                  gint             height,
                                  gpointer         user_data)
{
  gdk_pixbuf_loader_set_size(loader, width, height);
}


static void
libgnac_metadata_process_tag_image(const GstTagList *taglist, 
                                   gint              count,
                                   LibgnacMetadata  *md)
{
  gint i;
  for (i = 0; i < count; i++) 
  {
    const GValue *val;
    gint image_type;
    GstBuffer *buffer;
    GstCaps *caps;
    GstStructure *structure;

    val = gst_tag_list_get_value_index(taglist, GST_TAG_IMAGE, i);
      
    buffer = gst_value_get_buffer(val);
    g_return_if_fail(buffer);
      
    caps = GST_BUFFER_CAPS(buffer);
    g_return_if_fail(caps);

    structure = gst_caps_get_structure(caps, 0);
    g_return_if_fail(structure);

    gst_structure_get_enum(structure, "image-type", 
                    GST_TYPE_TAG_IMAGE_TYPE, &image_type);

    /* we are only interested in the front cover (but if there is only
     * one image that has an undefined type, we display it anyway) */
    if ((image_type == GST_TAG_IMAGE_TYPE_FRONT_COVER) ||
        (image_type == GST_TAG_IMAGE_TYPE_UNDEFINED && count == 1))
    {
      GdkPixbuf *pixbuf;
      GdkPixbufLoader *loader;
      GError *error = NULL;
      guint size;
      guint8 *data;
      GValue *newval;

      loader = gdk_pixbuf_loader_new();

      g_signal_connect(loader, "size-prepared", 
                      G_CALLBACK(libgnac_metadata_size_prepared_cb), NULL);
  
      data = GST_BUFFER_DATA(buffer);
      size = GST_BUFFER_SIZE(buffer);

      if (!gdk_pixbuf_loader_write(loader, data, size, &error)) 
      {
        libgnac_debug("Error writing data to pixbuf: %s\n", error->message);
        gdk_pixbuf_loader_close(loader, NULL);
        g_error_free(error);
        return;
      }

      if (!gdk_pixbuf_loader_close(loader, &error))
      {
        libgnac_debug("Error closing pixbuf loader: %s\n", error->message);
        g_error_free(error);
        return;
      }

      pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
      g_return_if_fail(pixbuf);

      newval = g_new0(GValue, 1);
      g_value_init(newval, G_TYPE_OBJECT);
      g_value_set_object(newval, pixbuf);

      /* we don't need the loader anymore */
      g_object_unref(G_OBJECT(loader));
      loader = NULL;

      g_hash_table_insert(md->priv->metadata, GST_TAG_IMAGE, newval);
    }
  }
}


static void
libgnac_metadata_process_tag(const GstTagList *taglist, 
                             const gchar      *tag_name, 
                             LibgnacMetadata  *md)
{
  gint count;
  GValue *newval;
  const GValue *val;
  GType type;
  
  count = gst_tag_list_get_tag_size(taglist, tag_name);
  if (count < 1) return;

  if (g_str_equal(tag_name, GST_TAG_IMAGE)) {
    libgnac_metadata_process_tag_image(taglist, count, md);
    return;
  }

  val = gst_tag_list_get_value_index(taglist, tag_name, 0);
  type = gst_tag_get_type(tag_name);
  newval = g_new0(GValue, 1);
  g_value_init(newval, type);
  if (!g_value_transform(val, newval)) 
  {
    libgnac_metadata_free_gvalue(newval);
    return;
  }

  if (type == G_TYPE_STRING) 
  {
    gchar *str;
    str = g_value_dup_string(newval);
    if (!g_utf8_validate(str, -1, NULL)) 
    {
      libgnac_info(_("Invalid UTF-8 tag: %s\n"), tag_name);
      g_free(str);
      libgnac_metadata_free_gvalue(newval);
      return;
    }
    str = g_strstrip(str);
    g_value_take_string(newval, str);
  }

  if (g_str_equal(tag_name, GST_TAG_BITRATE)) 
  {
    guint32 bitrate;
    bitrate = g_value_get_uint(newval);
    /* convert the bitrate from bps to kbps */
    g_value_set_uint(newval, bitrate/1000);
  }
  else if (g_str_equal(tag_name, GST_TAG_DURATION)) 
  {
    guint64 duration;
    duration = g_value_get_uint64(newval);
    /* convert the duration from ns to seconds */
    g_value_set_uint64(newval, duration/GST_SECOND);
  }
  else if (g_str_equal(tag_name, GST_TAG_DATE)) 
  {
    g_return_if_fail(GST_VALUE_HOLDS_DATE(newval));
    const GDate *date = gst_value_get_date(newval);
    if (!date) return;
    GDateYear year = g_date_get_year(date);
    g_value_unset(newval);
    g_value_init(newval, G_TYPE_UINT);
    g_value_set_uint(newval, year);
  }

  g_hash_table_insert(md->priv->metadata, (gpointer) tag_name, newval);
}


static void 
libgnac_metadata_on_message_application(GstBus          *bus,
                                        GstMessage      *message,
                                        LibgnacMetadata *md)
{
  libgnac_metadata_destroy_pipeline(md);
  md->priv->eos = TRUE;
}


static void
libgnac_metadata_on_message_eos(GstBus          *bus,
                                GstMessage      *message,
                                LibgnacMetadata *md)
{
  md->priv->eos = TRUE;
}


static void
libgnac_metadata_on_message_error(GstBus          *bus,
                                  GstMessage      *message,
                                  LibgnacMetadata *md)
{
  GError *error;
  gchar *debug;
  gst_message_parse_error(message, &error, &debug);
  libgnac_debug("Error on metadata reading: %s\n", error->message);

  md->priv->eos = TRUE; 
}


static void
libgnac_metadata_on_message_tag(GstBus          *bus,
                                GstMessage      *message,
                                LibgnacMetadata *md)
{
  GstTagList *taglist;
  gst_message_parse_tag(message, &taglist);
  if (GST_IS_TAG_LIST(taglist)) {
    gst_tag_list_foreach(taglist, 
                    (GstTagForeachFunc)libgnac_metadata_process_tag, md);
    gst_tag_list_free(taglist);
  }
}


static gboolean
libgnac_metadata_bus_handler(GstBus           *bus, 
                             GstMessage       *message, 
                             LibgnacMetadata  *md)
{
  g_return_val_if_fail(md, FALSE);

  switch (GST_MESSAGE_TYPE(message)) 
  {
    case GST_MESSAGE_EOS: 
      libgnac_metadata_on_message_eos(bus, message, md);
      break;
    case GST_MESSAGE_ERROR: 
      libgnac_metadata_on_message_error(bus, message, md);
      break;
    case GST_MESSAGE_TAG:
      libgnac_metadata_on_message_tag(bus, message, md);
      break;
    case GST_MESSAGE_APPLICATION:
      libgnac_metadata_on_message_application(bus, message, md);
      break;
    default:
      break;
  }

  return FALSE;
}


static GstBusSyncReply
libgnac_metadata_bus_sync_handler(GstBus          *bus, 
                                  GstMessage      *message, 
                                  LibgnacMetadata *md)
{
  libgnac_metadata_bus_handler(bus, message, md);
  gst_message_unref(message);
  return GST_BUS_DROP;
}


static void
libgnac_metadata_new_decoded_pad_cb(GstElement      *element,
                                    GstPad          *pad,
                                    gboolean         last,
                                    LibgnacMetadata *md)
{
  GstCaps *caps;
  GstPad  *sink_pad;
  GstStructure *structure;
  const gchar *mimetype;

  sink_pad = gst_element_get_pad(md->priv->sink, "sink");
  if (GST_PAD_IS_LINKED(sink_pad)) 
  {
    g_object_unref(sink_pad);
    return;
  }
  gst_pad_link(pad, sink_pad);

  caps = gst_pad_get_caps(pad);
  if (gst_caps_is_empty(caps) || gst_caps_is_any(caps)) 
  {
    gst_caps_unref(caps);
    g_object_unref(sink_pad);
    return;
  }

  structure = gst_caps_get_structure(caps, 0);
  mimetype = gst_structure_get_name(structure);

  if (!g_strrstr(mimetype, "audio")) 
  {
    gst_caps_unref(caps);
    gst_object_unref(sink_pad);
    return;
  }

  gst_caps_unref(caps);
  gst_object_unref(sink_pad);
}


static gboolean
libgnac_metadata_autoplug_continue_cb(GstElement      *decodebin,
                                      GstPad          *pad,
                                      GstCaps         *caps,
                                      LibgnacMetadata *md)
{
  gint rate, channels;
  GValue *rate_value, *channels_value;
  GstStructure *structure;

  g_return_val_if_fail(LIBGNAC_IS_METADATA(md), TRUE);
  
  if (!gst_caps_is_fixed(caps)) return TRUE;

  structure = gst_caps_get_structure(caps, 0);

  /* sample rate */
  if (gst_structure_get_int(structure, GNAC_TAG_RATE, &rate)) {
    rate_value = g_new0(GValue, 1);
    g_value_init(rate_value, G_TYPE_INT);
    g_value_set_int(rate_value, rate);
    g_hash_table_insert(md->priv->metadata, GNAC_TAG_RATE, rate_value);
  }

  /* channels */
  if (gst_structure_get_int(structure, GNAC_TAG_CHANNELS, &channels)) {
    channels_value = g_new0(GValue, 1);
    g_value_init(channels_value, G_TYPE_INT);
    g_value_set_int(channels_value, channels);
    g_hash_table_insert(md->priv->metadata, GNAC_TAG_CHANNELS, channels_value);
  }

  return TRUE;
}


static GstElement *
libgnac_metadata_make_pipeline_element(GstElement   *pipeline, 
                                       const gchar  *element, 
                                       GError      **error)
{
  GstElement *elem = gst_element_factory_make(element, element);
  if (!elem) {
    g_set_error(error, 
           LIBGNAC_ERROR, 
           LIBGNAC_ERROR_MISSING_PLUGIN, 
           _("Failed to create %s element"), 
           element);
    return NULL;
  }

  gst_bin_add(GST_BIN(pipeline), elem);
  return elem;
}        


static void
libgnac_metadata_get_file_info(GFile *uri, LibgnacMetadata *md)
{
  gchar     *path;
  GError    *error = NULL;
  GFileInfo *info;
  GValue    *file_size_value;
  GValue    *filename_value;
  GValue    *location_value;

  info = g_file_query_info(uri, 
        "standard::display-name,standard::size", G_FILE_QUERY_INFO_NONE, 
        NULL, &error);

  if (error) {
    g_clear_error(&error);
    return;
  }

  /* file size */
  file_size_value = g_new0(GValue, 1);
  g_value_init(file_size_value, G_TYPE_INT64);
  g_value_set_int64(file_size_value, g_file_info_get_size(info));
  g_hash_table_insert(md->priv->metadata, GNAC_TAG_FILE_SIZE, file_size_value);

  /* location */
  path = g_file_get_path(uri);
  location_value = g_new0(GValue, 1);
  g_value_init(location_value, G_TYPE_STRING);
  g_value_set_string(location_value, path);
  g_hash_table_insert(md->priv->metadata, GST_TAG_LOCATION, location_value);

  /* filename */
  filename_value = g_new0(GValue, 1);
  g_value_init(filename_value, G_TYPE_STRING);
  g_value_set_string(filename_value, g_file_info_get_display_name(info));
  g_hash_table_insert(md->priv->metadata, GNAC_TAG_FILENAME, filename_value);  

  /* cleanup */
  g_free(path);
  g_object_unref(info);
}


LibgnacTags *
libgnac_metadata_extract(LibgnacMetadata  *md,
                         GFile            *uri,
                         GError          **error)
{
  g_return_val_if_fail(LIBGNAC_IS_METADATA(md), NULL);
  g_return_val_if_fail(uri, NULL);

  gchar       *string_uri;
  LibgnacTags *tags = NULL;
  
  string_uri = g_file_get_uri(uri);
  tags = g_hash_table_lookup(md->priv->tags_table, string_uri);

  if (!tags) 
  {
    GstElement *pipeline  = NULL;
    GstElement *source    = NULL;
    GstElement *decodebin = NULL;
    GstBus *bus;

    /* make sure we start with a clean base */
    libgnac_metadata_reset(md);

    md->priv->uri = g_file_dup(uri);

    md->priv->metadata = g_hash_table_new_full(g_str_hash, g_str_equal,
        NULL, (GDestroyNotify) libgnac_metadata_free_gvalue);

    /* store file-size, location and filename */
    libgnac_metadata_get_file_info(uri, md);

    pipeline = gst_pipeline_new("pipeline");
    g_return_val_if_fail(pipeline, NULL);

    source = libgnac_metadata_make_pipeline_element(pipeline, "giosrc", error);
    g_return_val_if_fail(source, NULL);
    g_object_set(G_OBJECT(source), "file", md->priv->uri, NULL);

    decodebin = libgnac_metadata_make_pipeline_element(pipeline, "decodebin2", error);
    g_return_val_if_fail(decodebin, NULL);

    md->priv->sink = libgnac_metadata_make_pipeline_element(pipeline, "fakesink", error);
    g_return_val_if_fail(md->priv->sink, NULL);

    // XXX find out how to get rate/channels using decodebin instead of decodebin2...
    g_signal_connect(decodebin, "autoplug-continue", 
        G_CALLBACK(libgnac_metadata_autoplug_continue_cb), md);
    g_signal_connect_object(decodebin, "new-decoded-pad", 
        G_CALLBACK(libgnac_metadata_new_decoded_pad_cb), md, 0);

    gst_element_link(source, decodebin);
    md->priv->pipeline = pipeline;

    bus = gst_element_get_bus(GST_ELEMENT(pipeline));
    gst_bus_set_sync_handler(bus, (GstBusSyncHandler)libgnac_metadata_bus_sync_handler, md);
    gst_object_unref(bus);
    GstStateChangeReturn state_return;
    gint timeout_cycle;
    state_return = gst_element_set_state(md->priv->pipeline, GST_STATE_PAUSED);
    for (timeout_cycle = 0; timeout_cycle < 5; timeout_cycle++) {
      GstState current_state;
      if (state_return != GST_STATE_CHANGE_ASYNC || md->priv->eos) {
        break;
      }
      state_return = gst_element_get_state(md->priv->pipeline, &current_state, NULL, GST_SECOND);
    }

    if ((state_return == GST_STATE_CHANGE_FAILURE) ||
        (state_return == GST_STATE_CHANGE_ASYNC && !md->priv->eos)) 
    {
      libgnac_debug("An error occured while extracting metadata from file %s", 
                      string_uri);
      // XXX something went wrong...
      // what do we do?
    }
    else if (state_return == GST_STATE_CHANGE_SUCCESS) {
      if (g_hash_table_lookup(md->priv->metadata, GST_TAG_DURATION) == NULL) {
        GstFormat format = GST_FORMAT_TIME;
        gint64 length;
        GValue *newval;

        if (gst_element_query_duration(md->priv->sink, &format, &length)) {
          g_assert(format == GST_FORMAT_TIME);
          newval = g_new0(GValue, 1);
          g_value_init(newval, G_TYPE_UINT64);
          g_value_set_uint64(newval, length/GST_SECOND);
          g_hash_table_insert(md->priv->metadata, GST_TAG_DURATION, newval);
        } 
      }
    }
    else if (state_return != GST_STATE_CHANGE_FAILURE) {
      GstBus *message_bus = gst_element_get_bus(md->priv->pipeline);
      if (message_bus) {
        gst_bus_post(message_bus, gst_message_new_application(GST_OBJECT(md->priv->pipeline), NULL));
        gst_object_unref(message_bus);
      }
      while (!md->priv->eos);
    }
    
    libgnac_metadata_destroy_pipeline(md);
    g_hash_table_insert(md->priv->tags_table, g_strdup(string_uri), md->priv->metadata);
    tags = md->priv->metadata;

    // TODO We don't need to save anything in metadata anymore
    // put it to NULL ?
    // The Metadata will be freed when the hash table will be destroyed
  }

  g_free(string_uri);

  return tags;
}


void
libgnac_metadata_tags_free(LibgnacTags *tags)
{
  if (tags) {
    g_hash_table_destroy((GHashTable*)tags);
  }
}


gboolean
libgnac_metadata_tag_exists(LibgnacTags *tags,
                            const gchar *name)
{
  return libgnac_metadata_tags_exist(tags, name, NULL);
}


gboolean
libgnac_metadata_tags_exist(LibgnacTags *tags, ...)
{
  gboolean missing = FALSE;
  gchar *name = NULL;
  va_list names;

  va_start(names, tags);

  while ((name = va_arg(names, gchar*)) && !missing)
  {
    if (!g_hash_table_lookup(tags, name)) missing = TRUE;
  }

  va_end(names);

  return !missing;
}


LibgnacTags *
libgnac_metadata_get_dummy_tags(void)
{
  if (!dummy_tags)
  {
    dummy_tags = g_hash_table_new_full(g_str_hash, 
          g_str_equal, 
          g_free, 
          (GDestroyNotify)libgnac_metadata_tags_free);

    GValue *filename_value;
    filename_value = g_new0(GValue, 1);
    g_value_init(filename_value, G_TYPE_STRING);
    g_value_set_static_string(filename_value, _("filename.ext"));
    g_hash_table_insert(dummy_tags, GNAC_TAG_FILENAME, filename_value);

    GValue *album_value;
    album_value = g_new0(GValue, 1);
    g_value_init(album_value, G_TYPE_STRING);
    g_value_set_static_string(album_value, "Today!");
    g_hash_table_insert(dummy_tags, GST_TAG_ALBUM, album_value);
    
    GValue *artist_value;
    artist_value = g_new0(GValue, 1);
    g_value_init(artist_value, G_TYPE_STRING);
    g_value_set_static_string(artist_value, "The Beach Boys");
    g_hash_table_insert(dummy_tags, GST_TAG_ARTIST, artist_value);

    GValue *title_value;
    title_value = g_new0(GValue, 1);
    g_value_init(title_value, G_TYPE_STRING);
    g_value_set_static_string(title_value, "Do You Wanna Dance?");
    g_hash_table_insert(dummy_tags, GST_TAG_TITLE, title_value);

    GValue *genre_value;
    genre_value = g_new0(GValue, 1);
    g_value_init(genre_value, G_TYPE_STRING);
    g_value_set_static_string(genre_value, "Rock");
    g_hash_table_insert(dummy_tags, GST_TAG_GENRE, genre_value);

    GValue *comment_value;
    comment_value = g_new0(GValue, 1);
    g_value_init(comment_value, G_TYPE_STRING);
    g_value_set_static_string(comment_value, _("Converted by Gnac"));
    g_hash_table_insert(dummy_tags, GST_TAG_COMMENT, comment_value);

    GValue *date_value;
    date_value = g_new0(GValue, 1);
    g_value_init(date_value, G_TYPE_UINT);
    g_value_set_uint(date_value, 1965);
    g_hash_table_insert(dummy_tags, GST_TAG_DATE, date_value);

    GValue *track_count_value;
    track_count_value = g_new0(GValue, 1);
    g_value_init(track_count_value, G_TYPE_UINT);
    g_value_set_uint(track_count_value, 6);
    g_hash_table_insert(dummy_tags, GST_TAG_TRACK_COUNT, track_count_value);

    GValue *track_number_value;
    track_number_value = g_new0(GValue, 1);
    g_value_init(track_number_value, G_TYPE_UINT);
    g_value_set_uint(track_number_value, 1);
    g_hash_table_insert(dummy_tags, GST_TAG_TRACK_NUMBER, track_number_value);

    GValue *disc_number_value;
    disc_number_value = g_new0(GValue, 1);
    g_value_init(disc_number_value, G_TYPE_UINT);
    g_value_set_uint(disc_number_value, 1);
    g_hash_table_insert(dummy_tags, GST_TAG_ALBUM_VOLUME_NUMBER, 
                            disc_number_value);

    GValue *disc_count_value;
    disc_count_value = g_new0(GValue, 1);
    g_value_init(disc_count_value, G_TYPE_UINT);
    g_value_set_uint(disc_count_value, 2);
    g_hash_table_insert(dummy_tags, GST_TAG_ALBUM_VOLUME_COUNT, 
                            disc_count_value);
  }

  return dummy_tags;
}
