/*
 * $Id: gnac-main.c 380 2009-07-27 12:07:50Z bdupasqu $
 *
 * This file is part of GNAC - Gnome Audio Converter
 *
 * Copyright (C) 2007 - 2009 Gnac
 *    
 *    - DUPASQUIER  Benoit    <bdupasqu@users.sourceforge.net>
 *    - JOAQUIM     David     <daiji@users.sourceforge.net>
 *    - ROUX        Alexandre <alexandreroux@users.sourceforge.net>
 *
 * GNAC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GNAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNAC; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <gdk/gdkkeysyms.h>
#include <gio/gio.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gst/gst.h>
#include <gst/pbutils/pbutils.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "gnac.h"
#include "gnac-bars.h"
#include "gnac-file-list.h"
#include "gnac-gconf.h"
#include "gnac-ui.h"
#include "gnac-main.h"
#include "gnac-playlist.h"
#include "gnac-prefs.h"
#include "gnac-properties.h"
#include "gnac-utils.h"
#include "libgnac-converter.h"
#include "libgnac-error.h"
#include "libgnac-metadata.h"
#include "libgnac-debug.h"

#include "profiles/gnac-profiles.h"
#include "profiles/gnac-profiles-utils.h"


static GtkFileFilter    *current_file_filter;
static GtkFileFilter    *default_file_filter;
static GSList           *filters;
static gchar            *current_directory;
static LibgnacConverter *converter;
static gboolean          conversion_errors;
static guint64           prev_time_left;
static guint             timeout_id; 
//volatile?
static gboolean          continue_files_action;
static gboolean          quit_app = FALSE;
static GThread          *file_action_thread = NULL;
/* TODO add an option overwrite in gnac-prefs? */
static gboolean          overwrite = FALSE;
static gboolean          remember_overwrite = FALSE;
gint                     nb_files_total;
guint                    nb_files_added;

GnacState                state;
GnacState                prev_state;
LibgnacMetadata         *metadata;

#ifdef GNOME_ENABLE_DEBUG
  static const gchar *states[] = {
    "GNAC_AUDIO_EMPTY_STATE",
    "GNAC_AUDIO_FILE_ACTION_STATE",
    "GNAC_AUDIO_READY_STATE",
    "GNAC_AUDIO_CLEAR_STATE",
    "GNAC_AUDIO_CONVERT_STATE",
    "GNAC_AUDIO_PAUSED_STATE"
  };
#endif


static void
gnac_change_state(GnacState new_state)
{
  
#ifdef GNOME_ENABLE_DEBUG
  libgnac_debug("State changed to %s", states[new_state]);
#endif

  prev_state = state;
  state = new_state;
  
  g_object_get(G_OBJECT(converter), "nb-files", &nb_files_total, NULL);

  gnac_file_list_update_cursor();

  switch (state) {
    case GNAC_AUDIO_EMPTY_STATE:
      gnac_bars_on_row_deleted();
      if (gnac_gconf_get_boolean(GNAC_GCONF_TRAY_ICON)) gnac_ui_hide_trayicon();
      gnac_ui_show_progress(FALSE);
      gnac_ui_show_pause(FALSE);
      // TODO this activates clear button
      gnac_bars_on_convert_stop();
      if (timeout_id != 0) {
        g_source_remove(timeout_id);
        timeout_id = 0;
      }
      gnac_ui_set_progress_fraction(0.0f);
      gnac_ui_set_progress_text("");
      prev_time_left = -1;
      gnac_utils_moving_avg_reset();
      gnac_bars_activate_clear(FALSE);
    break;

    case GNAC_AUDIO_FILE_ACTION_STATE:
      timeout_id = g_timeout_add(100, 
        (GSourceFunc)gnac_ui_pulse_progress, NULL);
      gnac_bars_on_add_files();
      gnac_ui_show_progress(TRUE);
    break;

    case GNAC_AUDIO_READY_STATE: 
      remember_overwrite = FALSE;
      gnac_bars_on_row_inserted();
      if (gnac_gconf_get_boolean(GNAC_GCONF_TRAY_ICON)) gnac_ui_hide_trayicon();
      gnac_ui_show_progress(FALSE);
      gnac_ui_show_pause(FALSE);
      gnac_bars_on_convert_stop();
      if (timeout_id != 0) {
        g_source_remove(timeout_id);
        timeout_id = 0;
      }
      gnac_utils_moving_avg_reset();
      gnac_ui_set_progress_fraction(0.0f);
      gnac_ui_set_progress_text("");
      prev_time_left = -1;
      gnac_utils_moving_avg_reset();
    break;

    case GNAC_AUDIO_CLEAR_STATE:
      // TODO do stg? 
    break;
  
    case GNAC_AUDIO_CONVERT_STATE:
      if (gnac_gconf_get_boolean(GNAC_GCONF_TRAY_ICON)) gnac_ui_show_trayicon();
      gnac_file_list_hide_visual_bar();
      gnac_ui_show_progress(TRUE);
      gnac_ui_show_pause(TRUE);
      gnac_bars_on_convert();
      gnac_bars_on_convert_resume();
    break;

    case GNAC_AUDIO_PAUSED_STATE:
      gnac_bars_on_convert_pause();
    break;

  default:
    /* does nothing */
   break;

  } 
}


static void
gnac_return_prev_state(void)
{
  gnac_change_state(prev_state);
}


static void
gnac_add(GFile *file)
{
  GFile *child;
  GFileInfo *current_info;
  GFileEnumerator *enumerator;
  GError *error = NULL;
  
  if (continue_files_action) {
    switch (g_file_query_file_type(file, 
      G_FILE_QUERY_INFO_NONE, NULL)) {

      case G_FILE_TYPE_REGULAR:
        gnac_add_file(file); 
      break;

      case G_FILE_TYPE_DIRECTORY:
        enumerator = g_file_enumerate_children(file, "standard::type,standard::name", 
          G_FILE_QUERY_INFO_NONE, 
          NULL, &error);

        if (error) {
          libgnac_debug("%s", error->message);
          g_clear_error(&error);
          return;
        }

        current_info = g_file_enumerator_next_file(enumerator, NULL, &error);
        if (error) {
          libgnac_debug("%s", error->message);
          g_clear_error(&error);
          return;
        }

        while (current_info)
        {
          child = g_file_get_child(file, g_file_info_get_name(current_info));
          gnac_add(child);
          g_object_unref(child);
          current_info = g_file_enumerator_next_file(enumerator, NULL, &error);
          if (error) {
            libgnac_debug("%s", error->message);
            g_clear_error(&error);
            return;
          }
        }

        g_file_enumerator_close(enumerator, NULL, &error);
        if (error) {
            libgnac_debug("%s", error->message);
            g_clear_error(&error);
          return;
        }
      break;

      case G_FILE_TYPE_UNKNOWN:
      case G_FILE_TYPE_SYMBOLIC_LINK:
      case G_FILE_TYPE_SPECIAL:
      case G_FILE_TYPE_SHORTCUT:
      case G_FILE_TYPE_MOUNTABLE:
      default:
        /* Nothing is done*/
      break;
    }
  }
}


static void
gnac_files_names_list_foreach(gpointer data, 
                              gpointer user_data)
{
  GFile *file;

  file = (GFile*) data;
  gnac_add(file);
  g_object_unref(file);
}


static void
gnac_add_files_thread_entry(gpointer data) 
{
  GSList *files;
  gchar *files_added_text;
  
  gdk_threads_enter();
  gnac_file_list_attach_default_model(FALSE);
  gnac_ui_push_status(_("Importing files..."));
  gdk_flush();
  gdk_threads_leave();
  files = (GSList*) data;
  nb_files_added = 0;
  g_slist_foreach(files, (GFunc) gnac_files_names_list_foreach, NULL);
  g_slist_free(files);
  if (!quit_app) {
    gdk_threads_enter();
    gnac_ui_set_progress_text("");
    gnac_file_list_attach_default_model(TRUE);
    gnac_change_state(GNAC_AUDIO_READY_STATE);
    files_added_text = g_strdup_printf(
      ngettext("%u file imported", "%u files imported", nb_files_added), nb_files_added);
    gnac_ui_push_status(files_added_text);
    g_free(files_added_text);
    file_action_thread = NULL;
    gdk_flush();
    gdk_threads_leave();
  }
}


//The GList is freed by the thread itself.
static void
gnac_add_files(GSList *files)
{
  GError *error = NULL;
  continue_files_action = TRUE;
  gnac_change_state(GNAC_AUDIO_FILE_ACTION_STATE);
  file_action_thread = g_thread_create((GThreadFunc) gnac_add_files_thread_entry,
    files, TRUE, &error);
  if (error) {
    libgnac_warning(_("Impossible to launch thread to add files: %s\n"),
      error->message);
    g_error_free(error); 
    gnac_return_prev_state();
  }
}


static void
gnac_file_chooser_foreach(gpointer data, 
                          gpointer user_data)
{
  GFile *file;
  GSList **list;
  
  list = (GSList**) user_data;
  file = g_file_new_for_uri((gchar*)data);
  *list = g_slist_append(*list, file);
  g_free(data);
}


static void
gnac_init_filters(void)
{
  filters = (GSList*) NULL;
  
  const gchar *lossy_mime[][2] = {
    { "audio/mpeg"            , _("MPEG files (*.mp3)")                 }, 
    { "audio/mp4"             , _("MPEG-4 files (*.aac, *.m4a, *.mp4)") },
    { "audio/x-musepack"      , _("Musepack files (*.mpc)")             },
    { "application/ogg"       , _("Ogg Vorbis files (*.ogg)")           },
    { "audio/vnd.rn-realaudio", _("RealAudio files (*.ra)")             },
    { "audio/x-speex"         , _("Speex files (*.spx)")                },
    { "audio/x-ms-wma"        , _("Windows Media files (*.wma)")        },
    { NULL, NULL }
  };

  const gchar *lossless_mime[][2] = {
    /* XXX the Monkey's Audio plug-in has not yet been ported 
     * to gstreamer-0.10 */
    /*{ "audio/x-ape" , _("Ape files (*.ape)")   },*/
    { "audio/x-flac", _("Flac files (*.flac)") },
    { "audio/x-wav" , _("WAV files (*.wav)")   },
    { NULL, NULL }
  };

  const gchar *supported_mime_playlists[][2] = {
    { "audio/x-mpegurl"     , _("M3U playlists (*.m3u)")   },
    { "audio/x-scpls"       , _("PLS playlists (*.pls)")   },
    { "application/xspf+xml", _("XSPF playlists (*.xspf)") },
    { NULL, NULL }
  };

  GtkFileFilter *file_filter_all = gtk_file_filter_new();
  gtk_file_filter_add_pattern(file_filter_all, "*");
  gtk_file_filter_set_name(file_filter_all, _("All files (*.*)"));

  GtkFileFilter* file_filter_supported = gtk_file_filter_new();
  gtk_file_filter_set_name(file_filter_supported, _("All supported files"));

  /* all supported types */

  guint i = 0;
  while (lossless_mime[i][0] != NULL) {
    gtk_file_filter_add_mime_type(file_filter_supported, lossless_mime[i][0]);
    i++;
  }

  i = 0;
  while (lossy_mime[i][0] != NULL) {
    gtk_file_filter_add_mime_type(file_filter_supported, lossy_mime[i][0]);
    i++;
  }

  i = 0;
  GtkFileFilter *file_filter_playlists = gtk_file_filter_new();
  gtk_file_filter_set_name(file_filter_playlists, _("All playlists files"));
  while (supported_mime_playlists[i][0] != NULL) {
    gtk_file_filter_add_mime_type(file_filter_supported, supported_mime_playlists[i][0]);
    gtk_file_filter_add_mime_type(file_filter_playlists, supported_mime_playlists[i][0]);
    i++;
  }

  filters = g_slist_append(filters, (gpointer) file_filter_all);
  filters = g_slist_append(filters, (gpointer) file_filter_supported);
  filters = g_slist_append(filters, (gpointer) file_filter_playlists);

  /* Lossless types */

  i = 0;
  GtkFileFilter *file_filter_lossless = gtk_file_filter_new();
  gtk_file_filter_set_name(file_filter_lossless, _("Lossless codecs"));
  while (lossless_mime[i][0] != NULL) {
    gtk_file_filter_add_mime_type(file_filter_lossless, lossless_mime[i][0]);
    i++;
  }
  filters = g_slist_append(filters, (gpointer) file_filter_lossless);

  /* Lossy types */

  i = 0;
  GtkFileFilter *file_filter_lossy = gtk_file_filter_new();
  gtk_file_filter_set_name(file_filter_lossy, _("Lossy codecs"));
  while (lossy_mime[i][0] != NULL) {
    gtk_file_filter_add_mime_type(file_filter_lossy, lossy_mime[i][0]);
    i++;
  }
  filters = g_slist_append(filters, (gpointer) file_filter_lossy);

  /* Create filters for each kind of audio files */

  i = 0;
  GtkFileFilter *file_filter_tmp;
  while (lossless_mime[i][0] != NULL) {
    file_filter_tmp = gtk_file_filter_new();
    gtk_file_filter_add_mime_type(file_filter_tmp, lossless_mime[i][0]);
    gtk_file_filter_set_name(file_filter_tmp, lossless_mime[i][1]);
    filters = g_slist_append(filters, (gpointer) file_filter_tmp);
    i++;
  }

  i = 0;
  while (lossy_mime[i][0] != NULL) {
    file_filter_tmp = gtk_file_filter_new();
    gtk_file_filter_add_mime_type(file_filter_tmp, lossy_mime[i][0]);
    gtk_file_filter_set_name(file_filter_tmp, lossy_mime[i][1]);
    filters = g_slist_append(filters, (gpointer) file_filter_tmp);
    i++;
  }

  /* Create filters for playlists */

  i = 0;
  while (supported_mime_playlists[i][0] != NULL) {
    file_filter_tmp = gtk_file_filter_new();
    gtk_file_filter_add_mime_type(file_filter_tmp, supported_mime_playlists[i][0]);
    gtk_file_filter_set_name(file_filter_tmp, supported_mime_playlists[i][1]);
    filters = g_slist_append(filters, (gpointer) file_filter_tmp);
    i++;
  }

  GSList *next;
  for (next = filters; next; next = next->next) {
    g_object_ref(G_OBJECT(next->data)); 
  }
  current_file_filter = file_filter_supported;
  default_file_filter = file_filter_supported;
}


static void
gnac_add_filters(GtkFileChooser *file_chooser)
{
  GSList *next;

  for (next = filters; next; next = next->next) {
    gtk_file_chooser_add_filter (file_chooser, GTK_FILE_FILTER(next->data));
  }

  /* we want to keep track of the chosen file filter 
   * (Supported files at the beginning) */
  gtk_file_chooser_set_filter(file_chooser, current_file_filter);
}


static void
gnac_unref_filters(void)
{
  GSList *next;
  for (next = filters; next; next = next->next) {
     g_object_ref_sink(G_OBJECT(next->data)); 
     g_object_unref(next->data);
  }
  g_slist_free(filters);
}


void
gnac_add_file(GFile *file)
{
  GError *error = NULL;
  GFileInfo *info;
  GtkFileFilterInfo filter_info;

  //libgnac_info(_("Add file %s"), g_file_get_uri(file));

  info = g_file_query_info(file, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
          G_FILE_QUERY_INFO_NONE, NULL, &error);
  if (error) {
    /* Error adding file */
    libgnac_debug("Error adding file");
    g_clear_error(&error);
    return;
  }
  
  filter_info.contains =  GTK_FILE_FILTER_MIME_TYPE;
  filter_info.mime_type = g_file_info_get_content_type(info);

  if (filter_info.mime_type != NULL) {
    /* Playlists */
    if (g_ascii_strcasecmp(filter_info.mime_type, "audio/x-mpegurl") == 0) {
      gnac_playlist_parse_m3u(file);
    } else if (g_ascii_strcasecmp(filter_info.mime_type, "audio/x-scpls") == 0) {
      gnac_playlist_parse_pls(file);
    } else if (g_ascii_strcasecmp(filter_info.mime_type, "application/xspf+xml") == 0) {
      gnac_playlist_parse_xspf(file); 

    /* Filter the files */
    } else if (gtk_file_filter_filter(current_file_filter, &filter_info)) {

      /* No duplicate files */
      libgnac_converter_add(converter, file, &error);
    }
  }
  g_object_unref(info);
}


static void
gnac_about_url_hook(GtkAboutDialog *dialog,
                    const gchar    *url,
                    gpointer        user_data)
{
  GError *error = NULL;
  if (!gtk_show_uri(gtk_widget_get_screen(GTK_WIDGET(dialog)), 
                    url, gtk_get_current_event_time(), &error))
  {
    g_printerr(_("Could not open link %s: %s\n"), url, error->message);
    g_error_free(error);
  }
}


static void
gnac_about_email_hook(GtkAboutDialog *dialog,
                      const gchar    *email_address,
                      gpointer        user_data)
{
  gchar *uri;
  uri = g_strdup_printf("mailto:%s", email_address);
  gnac_about_url_hook(dialog, uri, user_data);
  g_free(uri);
}


static gboolean
gnac_confirm_exit(void)
{
  gint       response;
  GtkWidget *dialog;
  
  dialog = gtk_message_dialog_new(
      GTK_WINDOW(gnac_ui_get_widget("main_window")),
      GTK_DIALOG_MODAL,
      GTK_MESSAGE_WARNING,
      GTK_BUTTONS_YES_NO,
      _("A conversion is currently running...\n"
        "Are you sure you want to quit?"));

  gtk_window_set_title(GTK_WINDOW(dialog), APPNAME);

  response = gtk_dialog_run(GTK_DIALOG(dialog));

  gtk_widget_destroy(dialog);

  return (response == GTK_RESPONSE_YES);
}


gboolean
gnac_on_ui_destroy_cb(GtkWidget *widget, 
                      gpointer   data)
{
  GError *error = NULL;
  if (state == GNAC_AUDIO_CONVERT_STATE ||
      state == GNAC_AUDIO_PAUSED_STATE) 
  {
    if (!gnac_confirm_exit()) return TRUE;
    libgnac_converter_stop(converter, &error); 
  } else if (state == GNAC_AUDIO_FILE_ACTION_STATE) {
    GThread *thread = NULL;

    //TODO atomicity?
    if (file_action_thread != NULL) {
      thread = file_action_thread; 
      quit_app = TRUE;
      continue_files_action = FALSE;
      gnac_change_state(GNAC_AUDIO_READY_STATE);
    }
    if (thread != NULL) g_thread_join(thread);
  }

  g_free(current_directory);

  g_object_unref(converter);
  g_object_unref(metadata);

  gnac_unref_filters();

  gnac_properties_destroy();
  gnac_prefs_destroy();
  gnac_profiles_destroy();
  gnac_ui_destroy();

  gtk_main_quit();

  return FALSE;
}


static void
gnac_file_chooser_file_activated_cb(GtkFileChooser *chooser,
                                    gpointer        user_data)
{
  gnac_file_chooser_response(GTK_DIALOG(chooser), GTK_RESPONSE_NONE, NULL); 
}


static gboolean
gnac_file_chooser_key_press_event_cb(GtkWidget   *widget,
                                     GdkEventKey *event,
                                     gpointer     user_data)
{
  if (event->type == GDK_KEY_PRESS && event->keyval == GDK_Return)
  {
    gnac_file_chooser_response(GTK_DIALOG(widget), GTK_RESPONSE_NONE, NULL); 
    return TRUE;
  }
  return FALSE;
}


void
gnac_file_chooser_response(GtkDialog *dialog, 
                           gint       response, 
                           gpointer   user_data)
{
  GSList *list_path;
  GSList *list_files = NULL;

  GtkWidget *close_on_add_button;
  
  switch (response) {
    /* Add button */
    case GTK_RESPONSE_NONE:
      
      /* Set the current file filter */
      current_file_filter = gtk_file_chooser_get_filter(GTK_FILE_CHOOSER(dialog));
      list_path = gtk_file_chooser_get_uris(GTK_FILE_CHOOSER(dialog));

      /* Add the differents files */
      g_slist_foreach(list_path,(GFunc) gnac_file_chooser_foreach, (gpointer) &list_files);

      /* free the list */
      g_slist_free(list_path);
      
      /* add files */
      gnac_add_files(list_files);

      /* Do we have to close de file chooser? */
      close_on_add_button = gtk_file_chooser_get_extra_widget(GTK_FILE_CHOOSER(dialog));

      /* Get the current uri */
      g_free(current_directory);
      current_directory = gtk_file_chooser_get_current_folder_uri(GTK_FILE_CHOOSER(dialog));

      if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(close_on_add_button))) return;

      break;
     
    case GTK_RESPONSE_CLOSE:
    default: 
      break;
  }
  gtk_widget_destroy(GTK_WIDGET(dialog));
}


static gboolean
gnac_on_converter_overwrite_cb(LibgnacConverter *converter, 
                               gpointer          file,
                               gpointer          user_data)
{
  GFile *uri;
  GFileInfo *info;
  GError *error = NULL;

  uri = (GFile*) file;
  info = g_file_query_info(uri, G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
          G_FILE_QUERY_INFO_NONE, NULL, &error);

  if (!remember_overwrite) 
  {
    gint response;
    GtkWidget *dialog;
    GtkWidget *checkbox;

    dialog = gtk_message_dialog_new(
        GTK_WINDOW(gnac_ui_get_widget("main_window")),
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
        _("File %s already exists...\nOverwrite?"),
        g_file_info_get_display_name(info));
  
    gtk_window_set_title(GTK_WINDOW(dialog), APPNAME);

    checkbox = gtk_check_button_new_with_label(_("Remember my decision"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox), FALSE);
    gtk_box_pack_end(GTK_BOX(GTK_DIALOG(dialog)->vbox), checkbox, FALSE, FALSE, 0);
    gtk_widget_show(checkbox);

    gdk_threads_enter();
    response = gtk_dialog_run(GTK_DIALOG(dialog));
    gdk_threads_leave();

    remember_overwrite = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbox));

    gtk_widget_destroy(dialog);

    overwrite = (response == GTK_RESPONSE_YES);
  }

  g_object_unref(info);

  return overwrite;
}


static void
gnac_on_converter_progress_cb(LibgnacConverter *converter, 
                              guint             nb_converted,
                              gfloat            fraction,
                              guint64           time_left)
{
  if (state == GNAC_AUDIO_CONVERT_STATE) 
  {
    gint total;

    g_object_get(G_OBJECT(converter), "nb-files", &total, NULL);
    gnac_ui_push_status(g_strdup_printf(_("Converting file %d of %d"), 
                            nb_converted, total));
    gnac_ui_set_progress_fraction(fraction);

    if (prev_time_left != -1) 
    {
      gchar   *time_str;
      gchar   *time_str_label;
      guint64  dprogress;
      guint64  left;

      dprogress = prev_time_left - time_left;
      if (dprogress <= 0) return;
      
      left = ((time_left / dprogress) * PROGRESS_TIMEOUT) / 1000;

      time_str = gnac_utils_format_duration_for_display(
                      gnac_utils_moving_avg_add_sample(left));
      time_str_label = g_strdup_printf(_("%s left"), time_str);

      gnac_ui_set_progress_text(time_str_label);

      g_free(time_str_label);
      g_free(time_str);
    }

    prev_time_left = time_left;
  }
}


static void
gnac_on_converter_file_added_cb(LibgnacConverter *converter,
                                const gchar      *uri)
{
  //GError *error = NULL;
  //GFile *file;
  //GFileMonitor *monitor;
  ++nb_files_added;
  gnac_file_list_add_row(uri);

  // TODO hashtable to cancel monitor when file deletion on file-list
  /*
  file = g_file_new_for_uri(uri);
  monitor = g_file_monitor_file(file, 
      G_FILE_MONITOR_NONE,
      NULL, &error);
  if (error) {
    g_printerr(_("An error occured creating a file monitor"));
    g_clear_error(&error);
    return;
  }

  g_signal_connect(G_OBJECT(monitor),
      "changed", G_CALLBACK(gnac_on_file_monitor_changed_cb), NULL);

  g_object_unref(file);
  */
}


static void
gnac_on_converter_file_removed_cb(LibgnacConverter *converter,
                                  const gchar      *uri)
{
    gnac_file_list_remove_row(uri);
    if (!gnac_file_list_count_rows() > 0) {
      gnac_change_state(GNAC_AUDIO_EMPTY_STATE);
    }
}


static void
gnac_on_converter_file_started_cb(LibgnacConverter *converter,
                                  const gchar      *input,
                                  const gchar      *output)
{
  GError *err = NULL;

  gnac_file_list_remove_visual_error(input);

  gnac_file_list_select_uri_and_follow(input);

  if (err) {
    g_clear_error(&err);
    return;
  }

  if (gnac_gconf_get_boolean(GNAC_GCONF_TRAY_ICON)) {
    gnac_ui_trayicon_tooltip_update(input);
  }
}


static void
gnac_on_converter_file_completed_cb(LibgnacConverter *converter,
                                    const gchar      *uri)
{
  GFile *file;
  GError *error = NULL;
  
  /* Delete source if selected */
  if (gnac_gconf_get_boolean(GNAC_GCONF_CLEAR_SOURCE))
  {
    file = g_file_new_for_uri(uri);
    g_file_trash(file, NULL, &error);
    if (error)
    {
      if (g_error_matches(error, 
              G_IO_ERROR,
              G_IO_ERROR_NOT_SUPPORTED)) 
      {
        gint response;
        const gchar *filename;

        g_clear_error(&error);
        filename = gnac_utils_get_display_name(uri, &error);
        if (error) {
          g_clear_error(&error);
          g_object_unref(file);
          return;
        }

        response = gnac_ui_show_error_trash((gchar*)filename);

        switch (response) {
          case 1:
            g_file_delete(file, NULL, &error);
            if (error) {
              g_printerr(_("Error: %s\n"), error->message);
              g_clear_error(&error);
            }
          break;
          case GTK_RESPONSE_CANCEL:
          case GTK_RESPONSE_CLOSE:
          default:
          /* do nothing */
          break;
        }
      }
      g_object_unref(file);
      g_clear_error(&error);
    }
  }
}


static void
gnac_on_converter_files_cleared_cb(LibgnacConverter *converter)
{
  gnac_file_list_remove_all();
  gnac_change_state(GNAC_AUDIO_EMPTY_STATE);
  gnac_ui_push_status("");
}


static void
gnac_on_converter_completion_cb(LibgnacConverter *converter)
{
  gnac_change_state(GNAC_AUDIO_READY_STATE);
  if (conversion_errors) {
    gnac_ui_push_status(_("Conversion completed with errors"));
  } else {
    gnac_ui_push_status(_("Conversion completed!"));
  }
}


static void
gnac_on_converter_started_cb(LibgnacConverter *converter)
{
  gnac_change_state(GNAC_AUDIO_CONVERT_STATE);
  conversion_errors = FALSE;
}


static void
gnac_on_converter_stopped_cb(LibgnacConverter *converter)
{
  gnac_change_state(GNAC_AUDIO_READY_STATE);
  gnac_ui_push_status(_("Conversion stopped"));
}


static void
gnac_on_converter_error_cb(LibgnacConverter *converter,
                           const gchar      *uri,
                           const gchar      *msg,
                           GError           *error)
{
  if (g_error_matches(error, LIBGNAC_ERROR, LIBGNAC_ERROR_FILE_EXISTS))
  {
    gchar *text;
    text = g_strdup_printf(_("File not converted (%s)"), msg);
    gnac_file_list_visual_error(uri, text, GTK_STOCK_DIALOG_INFO);
    g_free(text);
  }
  else
  {
    conversion_errors = TRUE;
    gnac_file_list_visual_error(uri, msg, NULL);
    g_printerr(_("Gnac failed to convert file %s\n Error message: %s\n"), uri, msg);
  }
}


/* UI Callbacks */

void
gnac_on_ui_pause_cb(GtkWidget *widget,
                    gpointer   data)
{
  GError *error = NULL;

  if (state == GNAC_AUDIO_CONVERT_STATE)
  {
    gnac_change_state(GNAC_AUDIO_PAUSED_STATE);
    libgnac_converter_pause(converter, &error);
  }
  else
  {
    gnac_change_state(GNAC_AUDIO_CONVERT_STATE);
    libgnac_converter_resume(converter, &error);
  }

  if (error) 
  {
    /* do something with the error... */
    g_clear_error(&error);
  }
}

void
gnac_on_ui_convert_cb(GtkWidget *widget, 
                      gpointer   data)
{
  GError *error = NULL;

  switch(state) {
    case GNAC_AUDIO_CONVERT_STATE:
    case GNAC_AUDIO_PAUSED_STATE:
      libgnac_converter_stop(converter, &error);
    break;

    case GNAC_AUDIO_READY_STATE:
    {
      gchar *folder_hierarchy;
      gchar *folder_path;
      gchar *rename_pattern;
      folder_hierarchy = gnac_gconf_get_string(GNAC_GCONF_FOLDER_HIERARCHY_PATTERN),
      folder_path = gnac_gconf_get_string(GNAC_GCONF_DESTINATION_DIRECTORY);
      rename_pattern = gnac_gconf_get_string(GNAC_GCONF_RENAME_PATTERN_PATTERN);
      g_object_set(G_OBJECT(converter),
          "extension", gnac_profiles_get_extension(),
          "folder-hierarchy", folder_hierarchy,
          "folder-path", folder_path,
          "folder-type", gnac_gconf_get_int(GNAC_GCONF_FOLDER_TYPE), 
          "pipeline-description", gnac_profiles_get_pipeline(),
          "rename-pattern", rename_pattern,
          "strip-special", gnac_gconf_get_boolean(GNAC_GCONF_STRIP_SPECIAL),
          NULL);
      libgnac_converter_start(converter, &error);
      g_free(folder_path);
      g_free(folder_hierarchy);
      g_free(rename_pattern);
      break;
    }

    case GNAC_AUDIO_FILE_ACTION_STATE:
      continue_files_action = FALSE;
    break;

    default:
      // uha we have a problem
      g_return_if_reached();
    break;
  }
  if (error)
  {
    g_clear_error(&error);
  }
}


void
gnac_on_ui_profile_changed_cb(GtkComboBox *box,
                              gpointer     data)
{
  const gchar *profile_name = NULL;
  profile_name = gnac_profiles_get_name();
  gnac_gconf_set_string(GNAC_GCONF_LAST_USED_PROFILE, profile_name);
}


void
gnac_on_ui_add_cb(GtkWidget *widget, 
                  gpointer   data)
{
  GtkWidget *close_on_add_button;
  GtkWidget *file_chooser;
  
  file_chooser = gtk_file_chooser_dialog_new(_("Open a file..."),
      GTK_WINDOW(data),
      GTK_FILE_CHOOSER_ACTION_OPEN,
      GTK_STOCK_ADD, GTK_RESPONSE_NONE,
      GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
      NULL);

  gtk_window_set_type_hint(GTK_WINDOW(file_chooser), GDK_WINDOW_TYPE_HINT_DIALOG);

  gtk_file_chooser_set_current_folder_uri(GTK_FILE_CHOOSER(file_chooser), 
                  current_directory);
  gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(file_chooser), TRUE);

  close_on_add_button = gtk_check_button_new_with_label(_("Close dialog on add"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(close_on_add_button),TRUE);
  gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(file_chooser),
                  close_on_add_button);

  gnac_add_filters(GTK_FILE_CHOOSER(file_chooser));

  g_signal_connect(file_chooser, "response", 
                  G_CALLBACK(gnac_file_chooser_response), data);
  g_signal_connect(file_chooser, "key-press-event", 
                  G_CALLBACK(gnac_file_chooser_key_press_event_cb), data);
  g_signal_connect(file_chooser, "file-activated", 
                  G_CALLBACK(gnac_file_chooser_file_activated_cb), data);

  gtk_dialog_run(GTK_DIALOG(file_chooser));
}


static void
gnac_remove_track(gpointer data,
                  gpointer user_data)
{
  GError *error = NULL;
  GFile *file;
  GtkTreeRowReference *ref;
  gchar *uri = NULL;

  ref = (GtkTreeRowReference*) data;

  gnac_file_list_get(ref, &uri);
  file = g_file_new_for_uri(uri);
  //g_printerr(_("[UI] - Remove file %s\n"), uri);
  libgnac_converter_remove(converter, file, &error); 
  if (error)
  {
    g_printerr(_("Error: %s\n"), error->message);
    g_clear_error(&error);
  }

  g_object_unref(file);
  gtk_tree_row_reference_free(ref);
  g_free(uri);
}


void
gnac_on_ui_remove_cb(GtkWidget *widget, 
                     gpointer   data)
{
  GList *rows;

  rows = gnac_file_list_get_selected_rows();
  /* References are freed by gnac_remove_track*/
  g_list_foreach(rows, (GFunc) gnac_remove_track, NULL);
  g_list_free(rows);
}


void
gnac_on_ui_clear_cb(GtkWidget *widget, 
                    gpointer   data)
{
  GError *error = NULL;

  libgnac_converter_clear(converter);
  if (error)
  {
    g_printerr(_("Error: %s\n"), error->message);
    g_clear_error(&error);
  }
}


void
gnac_on_ui_help_cb(GtkWidget *widget, 
                   gpointer   data)
{
  GtkWidget *dialog;
  
  dialog = gtk_message_dialog_new(
      GTK_WINDOW(gnac_ui_get_widget("main_window")),
      GTK_DIALOG_DESTROY_WITH_PARENT,
      GTK_MESSAGE_ERROR,
      GTK_BUTTONS_CLOSE,
      _("No help available yet"));

  gtk_window_set_title(GTK_WINDOW(dialog), APPNAME);

  g_signal_connect_swapped(dialog, "response", 
      G_CALLBACK(gtk_widget_destroy), dialog);

  gtk_widget_show(dialog);
}


void
gnac_on_ui_properties_cb(GtkWidget *widget, 
                         gpointer   user_data)
{
  gnac_properties_window_show();
}


void
gnac_on_ui_drag_data_received_cb(GtkWidget        *widget, 
                                 GdkDragContext   *context, 
                                 gint              x, 
                                 gint              y,
                                 GtkSelectionData *selection_data, 
                                 guint             info, 
                                 guint             time, 
                                 gpointer          data)
{
  GFile  *file;
  GSList *list_uris = NULL;
  gchar **uris;
  gchar  *uri;
  gint index = 0;

  /* Disable any filtering for DnD. */
  current_file_filter = default_file_filter;

  uris = gtk_selection_data_get_uris(selection_data);

  if (uris == NULL) {
    gtk_drag_finish (context, FALSE, FALSE, time);
    return;
  }

  uri = uris[0];
  while (uri != NULL) {
    file = g_file_new_for_uri(uri);
    //gnac_add(file);
    //g_object_unref(file);
    list_uris = g_slist_append(list_uris, file);

    index = index + 1;
    uri = uris[index];
  }

  if (list_uris != NULL) gnac_add_files(list_uris);

  gtk_drag_finish(context, TRUE, FALSE, time);
  g_strfreev(uris);
}


/*static void
gnac_on_file_monitor_changed_cb(GFileMonitor      *monitor,
                                GFile             *file,
                                GFile             *other_file,
                                GFileMonitorEvent  event_type,
                                gpointer           user_data)
{
  GError *error = NULL;

  if (event_type == G_FILE_MONITOR_EVENT_DELETED) 
  {
    libgnac_converter_remove(converter, file, &error);
    g_object_unref(monitor);
    if (error) {
      g_error(_("An error occured on file monitor suppression"));
      g_clear_error(&error);
      return;
    }
  }
}*/


gint
main(gint    argc,
     gchar **argv)
{
  GOptionContext  *context;
  gchar **args_filenames = NULL;
  
  const GOptionEntry entries[] = {
    { G_OPTION_REMAINING, '\0', 0, G_OPTION_ARG_FILENAME_ARRAY, &args_filenames, NULL, NULL },
    { NULL }
  };

  if (!g_thread_supported()) g_thread_init(NULL);

  #ifdef ENABLE_NLS
    /* Internationalization */
    bindtextdomain(GETTEXT_PACKAGE, GNOMELOCALEDIR);
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
    textdomain(GETTEXT_PACKAGE);
  #endif /* ENABLE_NLS */

   /* make sure every window has its name and icon */
  g_set_application_name(APPNAME);
  gtk_window_set_default_icon_name("gnac");

  /* make the links clickable */
  gtk_about_dialog_set_url_hook(gnac_about_url_hook, NULL, NULL);
  gtk_about_dialog_set_email_hook(gnac_about_email_hook, NULL, NULL);

  /* Initialisation of libraries */
  gdk_threads_init();

  /* Parse command line arguments */
  context = g_option_context_new(_("[URI...] - Convert your audio files"));
  g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
  g_option_context_add_group(context, gst_init_get_option_group());
  g_option_context_add_group(context, gtk_get_option_group(TRUE));

  if (!g_option_context_parse(context, &argc, &argv, NULL))
  {
    gchar *help;
    help = g_option_context_get_help(context, TRUE, NULL);
    g_print("%s", help);
    g_free(help);
    exit(EXIT_FAILURE);
  }
  g_option_context_free(context);

  conversion_errors = FALSE;
  prev_time_left = -1;
  gnac_utils_moving_avg_init();
  state = prev_state = GNAC_AUDIO_EMPTY_STATE;

  /* Init gstreamer plugins helper*/
  gst_pb_utils_init();

  gnac_init_filters();
  gnac_gconf_init();
  gnac_profiles_init();

  current_directory = g_strdup(g_get_home_dir());

  /* Initialise converter and connect signals */
  metadata = LIBGNAC_METADATA(libgnac_metadata_new());
  converter = LIBGNAC_CONVERTER(libgnac_converter_new(metadata));
  g_signal_connect(converter, "allow-overwrite", 
      G_CALLBACK(gnac_on_converter_overwrite_cb), converter);
  g_signal_connect(converter, "progress", 
      G_CALLBACK(gnac_on_converter_progress_cb), converter);
  g_signal_connect(converter, "error", 
      G_CALLBACK(gnac_on_converter_error_cb), converter);
  g_signal_connect(converter, "completion", 
      G_CALLBACK(gnac_on_converter_completion_cb), converter);
  g_signal_connect(converter, "file-added", 
      G_CALLBACK(gnac_on_converter_file_added_cb), converter);
  g_signal_connect(converter, "file-removed", 
      G_CALLBACK(gnac_on_converter_file_removed_cb), converter);
  g_signal_connect(converter, "file-started", 
      G_CALLBACK(gnac_on_converter_file_started_cb), converter);
  g_signal_connect(converter, "file-completed", 
      G_CALLBACK(gnac_on_converter_file_completed_cb), converter);
  g_signal_connect(converter, "files-cleared", 
      G_CALLBACK(gnac_on_converter_files_cleared_cb), converter);
  g_signal_connect(converter, "started", 
      G_CALLBACK(gnac_on_converter_started_cb), converter);
  g_signal_connect(converter, "stopped", 
      G_CALLBACK(gnac_on_converter_stopped_cb), converter);
  gnac_ui_show();

  /* Check if arguments have been passed through the command line */
  GFile *arg_uri = NULL;
  GSList *list_files = NULL;
  if (args_filenames != NULL) 
  {
    gint i;
    gint len = g_strv_length(args_filenames);
    for (i = 0; i < len; i++) 
    {
      arg_uri = g_file_new_for_commandline_arg(args_filenames[i]);
      list_files = g_slist_append(list_files, arg_uri);
    }
    if (list_files != NULL) gnac_add_files(list_files);
    g_strfreev(args_filenames);
    args_filenames = NULL;
  }

  gdk_threads_enter();
  gtk_main();
  gdk_threads_leave();

  return EXIT_SUCCESS;
}
