/*
 * $Id: gnac-ui.c 379 2009-07-22 17:35:09Z bdupasqu $
 *
 * This file is part of GNAC - Gnome Audio Converter
 *
 * Copyright (C) 2007 - 2009 Gnac
 *    
 *    - DUPASQUIER  Benoit    <bdupasqu@users.sourceforge.net>
 *    - JOAQUIM     David     <daiji@users.sourceforge.net>
 *    - ROUX        Alexandre <alexandreroux@users.sourceforge.net>
 *
 * GNAC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GNAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNAC; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <glib/gi18n.h>
#include <glib/gprintf.h>

#include "gnac.h"
#include "gnac-bars.h"
#include "gnac-file-list.h"
#include "gnac-main.h"
#include "gnac-stock-items.h"
#include "gnac-ui.h"
#include "gnac-utils.h"
#include "gnac-gconf.h"
#include "profiles/gnac-profiles.h"

extern LibgnacMetadata *metadata;

static GtkBuilder     *gnac_main_builder = NULL;
static GtkStatusIcon  *trayicon = NULL;
static gchar          *status_msg = NULL;
static gchar          *tooltip_path = NULL;
static gchar          *progress_msg = NULL;
extern guint           nb_files_added;

static gint root_x;
static gint root_y;

static GtkWidget *audio_profile_chooser_combo = NULL;
static GtkWidget *audio_profile_edit_button = NULL;

static GtkTargetEntry target_list[] =  {
  { "text/uri-list", 0, TARGET_STRING }
};

static guint n_targets = G_N_ELEMENTS(target_list);

extern GnacState state;


void
gnac_ui_new(void)
{
  gchar     *current_profile;
  GtkWidget *file_list;
  GtkWidget *audio_profile_hbox;
  GError    *err = NULL;

  gnac_stock_items_init();

  gnac_main_builder = gtk_builder_new();

  gtk_builder_add_from_file(gnac_main_builder, PKGDATADIR "/gnac.xml", &err);
  if (err != NULL) 
  {
    g_printerr(_("Unable to read file: %s\n"), err->message);
    g_error_free (err);
  } 
  g_assert (err == NULL && gnac_main_builder != NULL);

  file_list = gnac_file_list_new();

  audio_profile_hbox = GTK_WIDGET(gtk_builder_get_object(gnac_main_builder, "audio_profile_hbox"));

  /* Add audio profile combo box */
  audio_profile_chooser_combo = gnac_profiles_get_widget(gnac_ui_get_widget("main_window"));

  g_signal_connect(G_OBJECT(audio_profile_chooser_combo), 
      "changed",
      G_CALLBACK(gnac_on_ui_profile_changed_cb), 
      NULL);

  current_profile = gnac_gconf_get_string(GNAC_GCONF_LAST_USED_PROFILE);
  gnac_profiles_set_current_profile(current_profile);
  g_free(current_profile);

  audio_profile_edit_button = gtk_button_new_from_stock(GTK_STOCK_EDIT);
  
  g_signal_connect(G_OBJECT(audio_profile_edit_button), 
      "clicked",
      G_CALLBACK(gnac_profiles_on_edit), 
      NULL);
  
  gtk_box_pack_start(GTK_BOX(audio_profile_hbox), 
      audio_profile_chooser_combo, 
      FALSE, TRUE, 0);

  gtk_box_pack_start(GTK_BOX(audio_profile_hbox), 
      audio_profile_edit_button, 
      FALSE, TRUE, 6);

  /* DnD portion code */
  
  gtk_drag_dest_set(file_list,
      GTK_DEST_DEFAULT_ALL,
      target_list,
      n_targets,
      GDK_ACTION_COPY);

  gtk_builder_connect_signals(gnac_main_builder, NULL);
}


void
gnac_ui_disable_profiles(void)
{
  gtk_widget_set_sensitive(audio_profile_chooser_combo, FALSE);   
  gtk_widget_set_sensitive(audio_profile_edit_button, FALSE);
}


void
gnac_ui_enable_profiles(void)
{
  gtk_widget_set_sensitive(audio_profile_chooser_combo, TRUE);   
  gtk_widget_set_sensitive(audio_profile_edit_button, TRUE);
}


void
gnac_ui_set_progress_fraction(gdouble fraction)
{
  GtkProgressBar *progress_bar;

  progress_bar = GTK_PROGRESS_BAR(
    gtk_builder_get_object(gnac_main_builder, "progressbar"));
  gtk_progress_bar_set_fraction(progress_bar, fraction);
}


void
gnac_ui_set_progress_text(gchar *text)
{
  GtkProgressBar *progress_bar;

  progress_msg = text;

  progress_bar = GTK_PROGRESS_BAR(
    gtk_builder_get_object(gnac_main_builder, "progressbar"));
  gtk_progress_bar_set_text(progress_bar, progress_msg);
}


void 
gnac_ui_show_progress(gboolean show)
{
  GtkWidget *progress_bar;

  progress_bar = GTK_WIDGET(
      gtk_builder_get_object(gnac_main_builder, "progressbar"));

  if (show) gtk_widget_show(progress_bar);
  else gtk_widget_hide(progress_bar);
}

gboolean 
gnac_ui_pulse_progress(void)
{
  GtkWidget *progress_bar;
  gchar *progress_text;

  if (state == GNAC_AUDIO_FILE_ACTION_STATE) {
    progress_bar = GTK_WIDGET(
        gtk_builder_get_object(gnac_main_builder, "progressbar"));
    
    gtk_progress_bar_pulse(GTK_PROGRESS_BAR(progress_bar));
    progress_text = g_strdup_printf(
      ngettext("%u file added", "%u files added", nb_files_added), nb_files_added);
    gnac_ui_set_progress_text(progress_text);
    g_free(progress_text);
    
    return TRUE;
  } else {
    return FALSE;
  }
}


void 
gnac_ui_show_pause(gboolean show)
{
  GtkWidget *button;

  button = GTK_WIDGET(
      gtk_builder_get_object(gnac_main_builder, "pause_button"));
  
  if (show) gtk_widget_show(button);
  else gtk_widget_hide(button);
}


/*
 * This function can be completed to have different contexts, libgnac, main-app.. 
 */
void
gnac_ui_push_status(const gchar *message)
{
  GtkStatusbar *status_bar;

  status_bar = GTK_STATUSBAR(
        gtk_builder_get_object(gnac_main_builder, "statusbar"));
  status_msg = g_strdup(message);
  gtk_statusbar_pop(status_bar, 0);
  gtk_statusbar_push(status_bar, 0, status_msg);
}


void
gnac_ui_append_status(const gchar *message)
{
  GtkStatusbar *status_bar;

  status_bar = GTK_STATUSBAR(
        gtk_builder_get_object(gnac_main_builder, "statusbar"));
  status_msg = g_strdup_printf("%s (%s)", status_msg, message);
  gtk_statusbar_pop(status_bar, 0);
  gtk_statusbar_push(status_bar, 0, status_msg);
}


void 
gnac_ui_show_popup_menu(GtkWidget      *treeview, 
                        GdkEventButton *event, 
                        gpointer        user_data)
{
  GtkWidget * popup_menu;

  popup_menu = GTK_WIDGET(gtk_builder_get_object(gnac_main_builder, 
                                                "main_popup"));

  gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, 
        (event != NULL) ? event->button : 0,
        gdk_event_get_time((GdkEvent*) event));
}


void
gnac_ui_show_about_dialog(void)
{
  GtkWidget *about_dialog;

  about_dialog = GTK_WIDGET(gtk_builder_get_object(gnac_main_builder, 
                                                   "aboutdialog"));
  gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(about_dialog), VERSION);
  gtk_widget_show_all(about_dialog);
}


GtkWidget *
gnac_ui_get_widget(const gchar *widget_name) 
{
  return GTK_WIDGET(gtk_builder_get_object(gnac_main_builder, widget_name));
}


GtkAction *
gnac_ui_get_action(const gchar *widget_name) 
{
  return GTK_ACTION(gtk_builder_get_object(gnac_main_builder, widget_name));
}


void 
gnac_ui_show(void)
{
  GtkWidget *main_window;
  GtkProgressBar *progress_bar;

  if (gnac_main_builder == NULL) 
  {
    gnac_ui_new();
  }

  progress_bar = GTK_PROGRESS_BAR(
    gtk_builder_get_object(gnac_main_builder, "progressbar"));

  main_window = GTK_WIDGET(gtk_builder_get_object(gnac_main_builder, "main_window"));
  gtk_widget_show_all(main_window);
  gnac_ui_show_progress(FALSE);
  gnac_ui_show_pause(FALSE);
  gnac_bars_activate_pause(FALSE);

}


void
gnac_ui_destroy(void)
{
  GtkWidget *main_window = NULL;
  GtkWidget *about_dialog = NULL;

  if (trayicon) {
    g_object_unref(trayicon);
    trayicon = NULL;
  }

  gnac_file_list_destroy();
  
  if (gnac_main_builder) {
    main_window = gnac_ui_get_widget("main_window");
    about_dialog = gnac_ui_get_widget("aboutdialog");
    g_object_unref(gnac_main_builder); 
    gnac_main_builder = NULL;
  }

  if (main_window) 
    gtk_widget_destroy(main_window);

  if (about_dialog)
    gtk_widget_destroy(about_dialog);
}


/* Tray icon */

/*
 * TODO
 * 
 * - make a nicer tooltip with, why not, a progressbar (c.f. audacious)
 *   (for a future version)
 *
 * - update label of Show/Hide depending on visibility or not of the window
 */

gboolean
gnac_ui_query_tooltip_cb(GtkStatusIcon *status_icon,
                         gint           x,
                         gint           y,
                         gboolean       keyboard_mode,
                         GtkTooltip    *tooltip,
                         gpointer       user_data)
{
  LibgnacTags *tags;
  GFile *tooltip_file;

  tooltip_file = g_file_new_for_uri(tooltip_path);
  tags = libgnac_metadata_extract(metadata, tooltip_file, NULL); 

  if (libgnac_metadata_tags_exist(tags, 
          GST_TAG_ARTIST, GST_TAG_ALBUM, GST_TAG_TITLE, NULL)) 
  {
    gchar *text;
    text = g_markup_printf_escaped(_("  <b>%s</b>\n"
              "  <span color=\"#888\">by</span> %s "
              "<span color=\"#888\">from</span> %s"), 
              g_value_get_string(LIBGNAC_METADATA_TAG_TITLE(tags)), 
              g_value_get_string(LIBGNAC_METADATA_TAG_ARTIST(tags)), 
              g_value_get_string(LIBGNAC_METADATA_TAG_ALBUM(tags)));
    gtk_tooltip_set_markup(tooltip, text);

    /* check wether we have a cover to display */
    if (libgnac_metadata_tag_exists(tags, GST_TAG_IMAGE)) 
    {
      GdkPixbuf *pixbuf;
      pixbuf = g_value_get_object(LIBGNAC_METADATA_TAG_IMAGE(tags));
      pixbuf = gnac_utils_scale_pixbuf(pixbuf, 80, 80);
      pixbuf = gnac_utils_add_border_to_pixbuf(pixbuf);
      gtk_tooltip_set_icon(tooltip, pixbuf);
    }
  } 
  else 
  {
    gtk_tooltip_set_text(tooltip, 
          gnac_utils_get_display_name(tooltip_path, NULL));
  }

  g_object_unref(tooltip_file);

  return TRUE;
}


void
gnac_ui_show_trayicon(void)
{
  GtkWidget *main_window;

  main_window = gnac_ui_get_widget("main_window");

  if (!trayicon) 
  {
    trayicon = gtk_status_icon_new_from_icon_name("gnac");
    gtk_status_icon_set_has_tooltip(trayicon, TRUE);
    
    g_signal_connect(G_OBJECT(trayicon), "query-tooltip",
        G_CALLBACK(gnac_ui_query_tooltip_cb), NULL);
    g_signal_connect(G_OBJECT(trayicon), "activate", 
        G_CALLBACK(gnac_ui_on_trayicon), NULL);
    g_signal_connect(G_OBJECT(trayicon), "popup-menu", 
        G_CALLBACK(gnac_ui_on_trayicon_popup), NULL);
  }
  
  gtk_window_get_position(GTK_WINDOW(main_window), &root_x, &root_y);
  gtk_status_icon_set_visible(trayicon, TRUE);
}


void
gnac_ui_hide_trayicon(void)
{  
  GtkWidget *main_window;
  gboolean window_displayed;

  main_window = gnac_ui_get_widget("main_window");
  g_object_get(main_window, "visible", &window_displayed, NULL);
  
  if (!window_displayed) 
  {
    gtk_widget_show_all(main_window);
    gtk_window_move(GTK_WINDOW(main_window), root_x, root_y);
  }

  if (!trayicon) 
  {
    trayicon = gtk_status_icon_new_from_icon_name("gnac");
    gtk_status_icon_set_has_tooltip(trayicon, TRUE);

    g_signal_connect(G_OBJECT(trayicon), "query-tooltip",
      G_CALLBACK(gnac_ui_query_tooltip_cb), NULL);
    g_signal_connect(G_OBJECT(trayicon), "activate", 
        G_CALLBACK(gnac_ui_on_trayicon), NULL);
    g_signal_connect(G_OBJECT(trayicon), "popup-menu", 
        G_CALLBACK(gnac_ui_on_trayicon_popup), NULL);
  }

  gtk_status_icon_set_visible(trayicon, FALSE);
}


void
gnac_ui_on_trayicon(GtkStatusIcon *trayicon,
                    gpointer       data)
{
  GtkWidget *main_window;
  gboolean window_displayed;

  main_window = gnac_ui_get_widget("main_window");
  g_object_get(main_window, "visible", &window_displayed, NULL);

  if (window_displayed) 
  {
    gtk_window_get_position(GTK_WINDOW(main_window), &root_x, &root_y);
    gtk_widget_hide_all(main_window);

  } else {
    gtk_widget_show_all(main_window);
    gtk_window_move(GTK_WINDOW(main_window), root_x, root_y);
  }
}


void
gnac_ui_on_trayicon_popup(GtkStatusIcon *trayicon,
                          guint          button,
                          guint          activate_time,
                          gpointer       data)
{
  gboolean   window_displayed;
  gchar     *label_text;
  GtkAction *show_action;
  GtkWidget *main_window;
  GtkWidget *trayicon_menu;
  
  /* get the widgets */
  main_window = gnac_ui_get_widget("main_window");
  show_action = gnac_ui_get_action("tray_show_hide_item");
  trayicon_menu = gnac_ui_get_widget("tray_popup");

  g_object_get(main_window, "visible", &window_displayed, NULL);
  label_text = g_strdup(window_displayed ? _("Hide") : _("Show"));

  /* update the menu and then display it */
  gtk_action_set_label(GTK_ACTION(show_action), label_text);
  gtk_menu_popup(GTK_MENU(trayicon_menu), NULL, NULL, NULL, NULL, 
        button, activate_time);

  /* cleanup */
  g_free(label_text);
}


void
gnac_ui_trayicon_tooltip_update(const gchar *tooltip)
{
  gchar *tmp = NULL;
  GdkDisplay *display;
  
  tmp = tooltip_path;
  tooltip_path = g_strdup(tooltip);
  display = gdk_display_get_default();
  if (display) gtk_tooltip_trigger_tooltip_query(display);
  g_free(tmp);
}


void
gnac_ui_trayicon_menu_stop(gpointer data, 
                           gpointer user_data)
{
  GtkWidget *main_window;
  gboolean window_displayed;

  main_window = gnac_ui_get_widget("main_window");
  g_object_get(main_window, "visible", &window_displayed, NULL);

  if (!window_displayed) {
    gtk_widget_show_all(main_window);
    gtk_window_move(GTK_WINDOW(main_window), root_x, root_y);
  }

  gnac_on_ui_convert_cb(NULL, NULL);
}


void
gnac_ui_trayicon_menu_pause(gpointer data,
                            gpointer user_data)
{
  gnac_on_ui_pause_cb(NULL, NULL);
}

/* Error dialogs */

gint
gnac_ui_show_error_trash(const gchar *filename)
{
  GtkWidget *main_window;
  GtkWidget *dialog;
  gint response;

  main_window = gnac_ui_get_widget("main_window");

  dialog = gtk_message_dialog_new(
      GTK_WINDOW(main_window),
      GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
      GTK_MESSAGE_QUESTION,
      GTK_BUTTONS_CANCEL,
      _("Cannot trash the file. Do you want to delete it immediately?"));
  gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_DELETE, 1);

  gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
   _("Cannot trash file %s"), filename);
  response = gtk_dialog_run(GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);

  return response;
}
