/*
 * $Id: gnac-profiles-xml-engine.c 240 2009-03-11 20:12:28Z alexandreroux $
 *
 * This file is part of GNAC - Gnome Audio Converter
 *
 * Copyright (C) 2007 - 2009 Gnac
 *    
 *    - DUPASQUIER  Benoit    <bdupasqu@users.sourceforge.net>
 *    - JOAQUIM     David     <daiji@users.sourceforge.net>
 *    - ROUX        Alexandre <alexandreroux@users.sourceforge.net>
 *
 * GNAC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GNAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNAC; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <glib/gstdio.h>
#include <stdio.h>
#include <string.h>
#include <libintl.h>


#include "gnac-profiles-xml-engine.h"

static xmlXPathObjectPtr
gnac_profiles_xml_engine_evaluate_query(const gchar *query,
                                        xmlXPathContextPtr xpath_context);

static gboolean
gnac_profiles_xml_engine_is_i12n_node(xmlNodePtr node);

XMLDoc *
gnac_profiles_xml_engine_load_doc_xpath(const gchar *filename)
{
  xmlDocPtr doc;
  xmlXPathContextPtr xpath;
  XMLDoc *dx = NULL;
  
  doc = xmlParseFile(filename);
  if (doc != NULL) {
    xpath = xmlXPathNewContext(doc);
    if (xpath != NULL) {
      dx = malloc(sizeof(XMLDoc));
      dx->doc = (void*) doc;
      dx->xpath_context = (void*) xpath;
    } else {
      xmlFreeDoc(doc);
    }
  }
  return dx;
}


gchar *
gnac_profiles_xml_engine_get_format_id(XMLDoc *dx)
{
  xmlXPathObjectPtr xpath_obj;
  const gchar *expr = "//profiles/profile";

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if ((nodes != NULL) && (nodes->nodeNr > 0)) {
      gchar *format_id;
      xmlNodePtr node;

      node = nodes->nodeTab[0];
      format_id = g_strdup((const gchar *) node->properties->children->content);
      xmlXPathFreeObject(xpath_obj);
      return format_id;
    }
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return NULL;
}

gchar *
gnac_profiles_xml_engine_get_text_node( XMLDoc *dx,
                                        const gchar *expr)
{
  xmlXPathObjectPtr xpath_obj; 

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if ((nodes != NULL) && (nodes->nodeNr > 0)) {
      xmlNodePtr node;
      //const xmlChar *content;

      node = nodes->nodeTab[0];
      if (node->children != NULL) {
        xmlXPathFreeObject(xpath_obj);
        return g_strdup(gettext((const gchar *)node->children->content));
      }
    }
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return NULL;
}

static gboolean
gnac_profiles_xml_engine_is_i12n_node(xmlNodePtr node)
{
  return !g_strcmp0((const gchar*) node->properties->name,"lang"); 
}

GList *
gnac_profiles_xml_engine_get_list_values( XMLDoc *dx,
                                          const gchar *expr)
{
  xmlXPathObjectPtr xpath_obj;

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if (nodes != NULL) {
      gint nb_element;
      gint i;
      GList *list = NULL;
      
      nb_element = nodes->nodeNr;
      for (i = 0; i<nb_element; ++i) {
        xmlNode *current_node = nodes->nodeTab[i];
        Value *v;
        
        if (!gnac_profiles_xml_engine_is_i12n_node(current_node)) {
          v = malloc(sizeof(Value));
          v->name = g_strdup(gettext((const gchar*) current_node->children->content));
          v->value = g_strdup((const gchar*) current_node->properties->children->content);
          list = g_list_append(list,v);
        }
        
      }
      xmlXPathFreeObject(xpath_obj);
      return list;
    }
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return NULL;
}

static xmlXPathObjectPtr
gnac_profiles_xml_engine_evaluate_query(const gchar *query,
                                        xmlXPathContextPtr xpath_context)
{
  const xmlChar* expr_cast;

  expr_cast = BAD_CAST query;
  return xmlXPathEvalExpression(expr_cast, xpath_context);
}

gint
gnac_profiles_xml_engine_get_combo_values(XMLDoc *dx,
                                          const gchar *expr,
                                          ComboValues *combo_values)
{
  xmlXPathObjectPtr xpath_obj; 

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if (nodes != NULL)
    {
      gint nb_element;
      gint i;
      gboolean is_combo = FALSE;
      xmlNode *values_first = NULL;
      xmlNode *values_last = NULL;
      gchar *default_value = NULL;
      gchar *variable_name = NULL;

      nb_element = nodes->nodeNr;
      for (i = 0; i<nb_element; ++i)
      {
        xmlNode *current_node = nodes->nodeTab[i];
        if (!g_strcmp0("variable-name",(const gchar*) current_node->name)) {
          variable_name = g_strdup((gchar*) current_node->children->content);
        } else if (!g_strcmp0("control-type",(const gchar*) current_node->name)) {
          is_combo =
            (!(g_strcmp0("combo",(const gchar*) current_node->children->content)));
        } else if (!g_strcmp0("default-value",(const gchar*) current_node->name)) {
          default_value = g_strdup((gchar*) current_node->children->content);
        } else if (!g_strcmp0("possible-values",(const gchar*) current_node->name)) {
          values_first = current_node->children->next;
          values_last = current_node->last;
        }
      }
      if (is_combo && values_first && values_last 
          && default_value && variable_name) {
        gboolean stop = FALSE;
        xmlNode *current_node = values_first;
        GList *values = NULL;
        GList *names = NULL;

        do {
          
          if (!gnac_profiles_xml_engine_is_i12n_node(current_node)) {
            values = g_list_append(values,
              g_strdup((gchar*)current_node->properties->children->content));

            names = g_list_append(names,
              g_strdup(gettext((const gchar*) current_node->children->content)));
          }
            
          current_node = current_node->next;
          if (current_node == values_last) {
            stop = TRUE;
          } else if (!g_strcmp0("text", (const gchar *) current_node->name)){
            current_node = current_node->next;
          }
        } while (!stop);
        
        combo_values->values = values;
        combo_values->names = names;
        combo_values->default_value = default_value;
        combo_values->variable_name = variable_name;
        xmlXPathFreeObject(xpath_obj);
        return 0;
      }
    }
  }
  
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return -1;
}

gint
gnac_profiles_xml_engine_get_slider_values(XMLDoc *dx,
                                          const gchar *expr,
                                          SliderValues *slider_values)
{
  xmlXPathObjectPtr xpath_obj; 

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if (nodes != NULL) {
      gint nb_element;
      gint i;
      gboolean is_slider = FALSE;
      gdouble min = 0.0;
      gdouble max = 0.0;
      gdouble step = 0.0;
      gdouble default_value = 0.0;
      gchar *variable_name = NULL;

      nb_element = nodes->nodeNr;
      for (i = 0; i<nb_element; ++i) {
        xmlNode *current_node = nodes->nodeTab[i];
        if (!g_strcmp0("variable-name",(const gchar*) current_node->name)) {
          variable_name = g_strdup((gchar*) current_node->children->content);
        } else if (!g_strcmp0("control-type",(const gchar*) current_node->name)){
          is_slider =
            (!(g_strcmp0("slider",(const gchar*) current_node->children->content)));
        } else if (!g_strcmp0("default-value",(const gchar*) current_node->name)) {
          default_value = g_ascii_strtod(
            (const gchar*) current_node->children->content, NULL);
        } else if (!g_strcmp0("min-value",(const gchar*) current_node->name)) {
          min = g_ascii_strtod(
            (const gchar*) current_node->children->content, NULL);
        } else if (!g_strcmp0("max-value",(const gchar*) current_node->name)) {
          max = g_ascii_strtod(
            (const gchar*) current_node->children->content, NULL);
        } else if (!g_strcmp0("step-value",(const gchar*) current_node->name)) {
          step = g_ascii_strtod(
            (const gchar*) current_node->children->content, NULL);
        }
      }
      if (is_slider && variable_name)
      {
        slider_values->min = min;
        slider_values->max = max;
        slider_values->step = step;
        slider_values->default_value = default_value;
        slider_values->variable_name = variable_name;
        xmlXPathFreeObject(xpath_obj);
        return 0;
      }
    }
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return -1;
}


gint
gnac_profiles_xml_engine_get_check_values(XMLDoc *dx,
                                          const gchar *expr,
                                          CheckValues *check_values)
{
  xmlXPathObjectPtr xpath_obj; 

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  if ((xpath_obj != NULL) && (!xpath_obj->boolval)) {
    xmlNodeSetPtr nodes = xpath_obj->nodesetval;
    
    if (nodes != NULL) {
      gint nb_element;
      gint i;
      gchar *variable_name = NULL;
      gboolean is_check = FALSE;
      gboolean default_value = FALSE;

      nb_element = nodes->nodeNr;
      for (i = 0; i<nb_element; ++i) {
        xmlNode *current_node = nodes->nodeTab[i];
        if (!g_strcmp0("variable-name",(const gchar*) current_node->name)) {
          variable_name = g_strdup((gchar*) current_node->children->content);
        } else if (!g_strcmp0("control-type",(const gchar*) current_node->name)){
          is_check =
            (!(g_strcmp0("check",(const gchar*) current_node->children->content)));
        } else if (!g_strcmp0("default-value",(const gchar*) current_node->name)) {
          default_value = 
            (!g_strcmp0((const gchar*) current_node->children->content,"true"));
        }
      }
      if (is_check && variable_name)
      {
        check_values->default_value = default_value;
        check_values->variable_name = variable_name;
        xmlXPathFreeObject(xpath_obj);
        return 0;
      }
    }
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
  return -1;
}

void
gnac_profiles_xml_engine_modify_values(XMLDoc *dx, const gchar *expr, ...)
{
  va_list ap;
  const gchar *value;
  xmlNodeSetPtr nodes;
  xmlXPathObjectPtr xpath_obj; 
  gint nbElement;
  gint i;

  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);

  nodes = xpath_obj->nodesetval;
  nbElement = nodes->nodeNr;
  
  va_start(ap,expr);
  value = va_arg(ap,const gchar *);
  for (i = nbElement-1; i >= 0; --i) {
    gchar *value_formated = gnac_profiles_xml_engine_format_text_to_xml(value);
    xmlNodeSetContent(nodes->nodeTab[i], BAD_CAST value_formated);
    g_free(value_formated);
    
    if (nodes->nodeTab[i]->type != XML_NAMESPACE_DECL)
	    nodes->nodeTab[i] = NULL;
    
    value = va_arg(ap, const gchar *);
  }
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
}

void
gnac_profiles_xml_engine_add_values(XMLDoc *dx, const gchar *expr, va_list ap)
{
  xmlNodeSetPtr nodes;
  xmlNodePtr format_node;
  xmlNodePtr node;
  xmlXPathObjectPtr xpath_obj; 
  const xmlChar *name;
  const gchar *value;
  const xmlChar *base_indent;
  xmlChar *properties_indent;
  
  xpath_obj = gnac_profiles_xml_engine_evaluate_query(expr, dx->xpath_context);
  
  nodes = xpath_obj->nodesetval;
  format_node = nodes->nodeTab[0]; 
  base_indent = format_node->prev->content;
  properties_indent = (xmlChar*) g_strconcat((const gchar*) base_indent,"  ",NULL);
  xmlAddChild(format_node, xmlNewText(BAD_CAST "  "));

  name = va_arg(ap, const xmlChar *);
  while (name != NULL) {

    value = va_arg(ap, const gchar *);
    
    node = xmlNewNode(format_node->ns,name);
    node = xmlAddChild(format_node, node);
    if (g_strcmp0("", value)) {
      gchar *value_formated = gnac_profiles_xml_engine_format_text_to_xml(value);
      xmlNodeSetContent(node, BAD_CAST value_formated);
      g_free(value_formated);
    }
    
    name = va_arg(ap, const xmlChar *);
    if (name != NULL) {
      xmlAddChild(format_node, xmlNewText(properties_indent));
    } else {
      xmlAddChild(format_node, xmlNewText(base_indent));
    }
  }
  va_end(ap);
  g_free(properties_indent);
  if (xpath_obj != NULL) xmlXPathFreeObject(xpath_obj);
}

gchar *
gnac_profiles_xml_engine_format_text_to_xml(const gchar *text) {
  if (text != NULL) {
    gchar *text_return = NULL;
    size_t text_size;
    const gchar *replacement;
    gchar *temp;
    int last_replacement = 0;
    int i;
    
    text_size = strlen(text);
    for (i = 0; i<text_size; ++i) {
      switch (text[i]) {
        case '&':
          replacement = "&amp;";
          break;
        case '<':
          replacement = "&lt;";
          break;
        case '>':
          replacement = "&gt;";
          break;
        default:
          replacement = NULL;
          break;
      }
      if (replacement != NULL) {
        temp = g_strndup(&(text[last_replacement]),i-last_replacement);
        if (text_return != NULL) {
          gchar *temp2;
          
          temp2 = text_return;
          text_return = g_strconcat(text_return, temp, replacement, NULL);
          g_free(temp2);
        } else {
          text_return = g_strconcat(temp, replacement, NULL);
        }
        g_free(temp);
        last_replacement = i+1;
      }
    }
    if (text_return == NULL) {
      return g_strdup(text);
    } else  {
      return text_return;
    }
  } else {
    return NULL;
  }
}

void
gnac_profiles_xml_engine_save_doc(XMLDoc *dx, const gchar *filename)
{
  gchar *xml_filename;

  xml_filename = g_strconcat(GNAC_SAVED_PROFILES_URL_WITH_EXT(filename), NULL);
  xmlSaveFormatFile(xml_filename,dx->doc, 0);

  g_free(xml_filename);
}

void
gnac_profiles_xml_engine_free_doc_xpath(XMLDoc *dx)
{
  xmlXPathFreeContext(dx->xpath_context); 
  xmlFreeDoc(dx->doc); 
  g_free(dx);
}

void
gnac_profiles_xml_engine_free_combo_values(ComboValues *values)
{
  g_list_foreach(values->values, (GFunc) g_free, NULL); 
  g_list_free(values->values);
  g_list_foreach(values->names, (GFunc) g_free, NULL); 
  g_list_free(values->names);
  g_free(values->default_value);
  g_free(values->variable_name);
  g_free(values);
}

void
gnac_profiles_xml_engine_free_slider_values(SliderValues *values)
{
  g_free(values->variable_name);
  g_free(values);
}

void
gnac_profiles_xml_engine_free_check_values(CheckValues *values)
{
  g_free(values->variable_name);
  g_free(values);
}
